# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Giorgio Fontanive (twodoorcoupe)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Post Tagging Actions"
PLUGIN_AUTHOR = "Giorgio Fontanive"
PLUGIN_DESCRIPTION = """
This plugin lets you set up actions that run with a context menu click. 
An action consists in a command line executed for each album or each track along
with a few options to tweak the behaviour. 
This can be used to run external programs and pass some variables to it. 
"""
PLUGIN_VERSION = "0.1"
PLUGIN_API_VERSIONS = ["2.10", "2.11"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"
PLUGIN_USER_GUIDE_URL = "https://github.com/metabrainz/picard-plugins/tree/2.0/plugins/post_tagging_actions/docs/guide.md"

from picard.album import Album
from picard.track import Track
from picard.ui.options import OptionsPage, register_options_page
from picard.ui.itemviews import BaseAction, register_album_action, register_track_action
from picard.ui import mainwindow
from picard import log, config
from picard.const import sys
from picard.util import thread
from picard.script import parser

from .options_post_tagging_actions import Ui_PostTaggingActions
from .actions_status import Ui_ActionsStatus
from PyQt5 import QtCore, QtWidgets, QtGui

from collections import namedtuple
from queue import PriorityQueue
from threading import Thread, Lock
from concurrent import futures
from os import path, cpu_count
import re
import shlex
import subprocess  # nosec B404
import time

WIDGET_UPDATE_INTERVAL = 0.5

# Additional special variables.
TRACK_SPECIAL_VARIABLES = {
    "filepath": lambda file: file,
    "folderpath": lambda file: path.dirname(file),  # pylint: disable=unnecessary-lambda
    "filename": lambda file: path.splitext(path.basename(file))[0],
    "filename_ext": lambda file: path.basename(file),  # pylint: disable=unnecessary-lambda
    "directory": lambda file: path.basename(path.dirname(file))
}
ALBUM_SPECIAL_VARIABLES = {
    "get_num_matched_tracks",
    "get_num_unmatched_files",
    "get_num_total_files",
    "get_num_unsaved_files",
    "is_complete",
    "is_modified"
}

# Settings.
CANCEL = "pta_cancel"
MAX_WORKERS = "pta_max_workers"
OPTIONS = ("pta_command", "pta_wait_for_exit", "pta_execute_for_tracks", "pta_refresh_tags")

Options = namedtuple("Options", ("variables", *[option[4:] for option in OPTIONS]))
Action = namedtuple("Action", ("commands", "album", "options"))
PriorityAction = namedtuple("PriorityAction", ("priority", "counter", "action"))
action_queue = PriorityQueue()
variables_pattern = re.compile(r'%.*?%')


class ActionLoader:
    """Adds actions to the execution queue.

    Attributes:
        action_options (list): Stores the actions' information loaded from the options page.
        action_counter (int): The count of actions that have been added to the queue, used for priority.
    """

    def __init__(self):
        self.action_options = []
        self.action_counter = 0
        self.load_actions()

    def _create_options(self, command, *other_options):
        """Finds the variables in the command and adds the options to the action options list.
        """
        variables = [parser.normalize_tagname(variable[1:-1]) for variable in variables_pattern.findall(command)]
        command = variables_pattern.sub('{}', command)
        options = Options(variables, command, *other_options)
        self.action_options.append(options)

    def _create_action(self, priority, commands, album, options):
        """Adds an action with the given parameters to the execution queue.

        If the os is not windows, the command is split as suggested by the subprocess
        module documentation.
        """
        if not sys.IS_WIN:
            commands = [shlex.split(command) for command in commands]
        action = Action(commands, album, options)
        priority_action = PriorityAction(priority, self.action_counter, action)
        action_queue.put(priority_action)
        self.action_counter += 1

    def _replace_variables(self, variables, item):
        """Returns a list where each variable is replaced with its value.

        Item is either an album or a track. For track special variables,
        it uses the path of the first file of the given item.
        If the variable is not found anywhere, it remains as in the original text.
        """
        values = []
        album = item.album if isinstance(item, Track) else item
        first_file_path = next(item.iterfiles()).filename
        for variable in variables:
            if variable in ALBUM_SPECIAL_VARIABLES:
                values.append(getattr(album, variable)())
            elif variable in TRACK_SPECIAL_VARIABLES:
                values.append(TRACK_SPECIAL_VARIABLES[variable](first_file_path))
            else:
                values.append(item.metadata.get(variable, f"%{variable}%"))
        return values

    def add_actions(self, album, tracks):
        """Adds one action to the execution queue for each tuple in the action options list.

        Actions meant to be executed once for each track are considered as a single
        action. This way, the other options are more consistent.
        """
        for priority, options in enumerate(self.action_options):
            if options.execute_for_tracks:
                values_list = [self._replace_variables(options.variables, track) for track in tracks]
            else:
                values_list = [self._replace_variables(options.variables, album)]
            commands = [options.command.format(*values) for values in values_list]
            self._create_action(priority, commands, album, options)

    def load_actions(self):
        """Loads the information from the options and saves it in the action options list.

        This gets called when the plugin is loaded or when the user saves the options.
        """
        self.action_options = []
        option_tuples = zip(*[config.setting[name] for name in OPTIONS])
        for option_tuple in option_tuples:
            command = option_tuple[0]
            other_options = [eval(option) for option in option_tuple[1:]]  # nosec B307
            self._create_options(command, *other_options)


class ActionRunner:
    """Runs actions in the execution queue.

    Attributes:
        action_thread_pool (ThreadPoolExecutor): Pool used to run processes with the subprocess module.
        refresh_tags_pool (ThreadPoolExecutor): Pool used to reload tags from files and refresh albums.
        worker (Thread): Worker thread that picks actions from the execution queue.
        update_widget (Thread): Thread that updates the number of pending actions in the status bar.
    """

    def __init__(self):
        self.action_thread_pool = futures.ThreadPoolExecutor(config.setting[MAX_WORKERS])
        self.refresh_tags_pool = futures.ThreadPoolExecutor(1)
        self.worker = Thread(target = self._execute)
        self.update_widget = Thread(target = self._update_widget)
        self.worker.start()

        self.keep_updating = True
        self.currently_executing = 0
        self.currently_executing_lock = Lock()
        self.status_widget = ActionsStatus()

        # This is used to register functions that run when the application is being closed.
        # The stop function makes the background threads return.
        tagger = QtCore.QCoreApplication.instance()
        tagger.register_cleanup(self.stop)

        # This checks whether the tagger has already created the main window.
        # It should happen only when the plugin is installed through the options menu,
        # otherwise the plugins are loaded before the main window is created.
        if hasattr(tagger, "window"):
            self._create_widget(tagger.window)
        else:
            # This is used to register functions that run after the main window has finished loading.
            mainwindow.register_ui_init(self._create_widget)

    def _create_widget(self, window):
        """Adds the pending actions widget to the right of the other icons in the statusbar.
        """
        window.statusBar().insertPermanentWidget(1, self.status_widget)
        self.update_widget.start()

    def _update_widget(self):
        """Updates the number of pending actions in the status bar at regular intervals.
        """
        while self.keep_updating:
            number_of_actions = action_queue.qsize() + self.currently_executing
            thread.to_main(self.status_widget.update_actions_count, number_of_actions)
            time.sleep(WIDGET_UPDATE_INTERVAL)

    def _refresh_tags(self, future_objects, album):
        """Reloads tags from the album's files and refreshes the album.

        First, it makes sure that the action has finished running. This is used for
        when an external process changes a file's tags.
        """
        futures.wait(future_objects, return_when = futures.ALL_COMPLETED)
        for file in album.iterfiles():
            file.set_pending()
            file.load(lambda file: None)
        thread.to_main(album.load, priority = True, refresh = True)

    def _run_process(self, command):
        """Runs the process and waits for it to finish.
        """
        process = subprocess.Popen(
            command,
            text = True,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE
        )  # nosec B603
        answer = process.communicate()
        if answer[0]:
            log.info("Action output:\n%s", answer[0])
        if answer[1]:
            log.error("Action error:\n%s", answer[1])

    def _update_executing_count(self, future_objects):
        """Decrements the count of executing actions once the given action finishes.
        """
        futures.wait(future_objects, return_when = futures.ALL_COMPLETED)
        with self.currently_executing_lock:
            self.currently_executing -= 1

    def _execute(self):
        """Takes actions from the execution queue and runs them.

        If it finds an action with priority -1, the loop stops. When the loop
        stops, both ThreadPoolExecutors are shutdown.
        """
        while True:
            priority_action = action_queue.get()
            if priority_action.priority == -1:
                break
            with self.currently_executing_lock:
                self.currently_executing += 1
            next_action = priority_action.action
            commands = next_action.commands
            future_objects = {self.action_thread_pool.submit(self._run_process, command) for command in commands}

            if next_action.options.wait_for_exit:
                futures.wait(future_objects, return_when = futures.ALL_COMPLETED)
            if next_action.options.refresh_tags:
                self.refresh_tags_pool.submit(self._refresh_tags, future_objects, next_action.album)
            self.refresh_tags_pool.submit(self._update_executing_count, future_objects)
            action_queue.task_done()

        self.action_thread_pool.shutdown(wait = False, cancel_futures = True)
        self.refresh_tags_pool.shutdown(wait = False, cancel_futures = True)

    def stop(self):
        """Makes the worker thread exit its loop.

        This gets called when Picard is closed. It waits for the processes that
        are still executing to finish before exiting.
        """
        if not config.setting[CANCEL]:
            action_queue.join()
        action_queue.put(PriorityAction(-1, -1, None))
        self.keep_updating = False
        if self.update_widget.is_alive():
            self.update_widget.join()
        self.worker.join()


class ExecuteAlbumActions(BaseAction):

    NAME = "Run actions for highlighted albums"

    def callback(self, objs):
        albums = {obj for obj in objs if isinstance(obj, Album)}
        for album in albums:
            action_loader.add_actions(album, album.tracks)


class ExecuteTrackActions(BaseAction):

    NAME = "Run actions for highlighted tracks"

    def callback(self, objs):
        tracks = {obj for obj in objs if isinstance(obj, Track)}
        albums = {track.album for track in tracks}
        for album in albums:
            album_tracks = tracks.intersection(album.tracks)
            action_loader.add_actions(album, album_tracks)


class PostTaggingActionsOptions(OptionsPage):
    """Options page found under the "plugins" page.
    """

    NAME = "post_tagging_actions"
    TITLE = "Post Tagging Actions"
    PARENT = "plugins"

    action_options = [config.ListOption("setting", name, []) for name in OPTIONS]
    options = [
        config.BoolOption("setting", CANCEL, True),
        config.IntOption("setting", MAX_WORKERS, min(32, cpu_count() + 4)),
        *action_options
    ]

    def __init__(self, parent = None):
        super(PostTaggingActionsOptions, self).__init__(parent)
        self.ui = Ui_PostTaggingActions()
        self.ui.setupUi(self)
        self._reset_ui()

        header = self.ui.table.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        for column in range(1, header.count()):
            header.setSectionResizeMode(column, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

        self.ui.add_file_path.clicked.connect(self._open_file_dialog)
        self.ui.add_action.clicked.connect(self._add_action_to_table)
        self.ui.remove_action.clicked.connect(self._remove_action_from_table)
        self.ui.up.clicked.connect(self._move_action_up)
        self.ui.down.clicked.connect(self._move_action_down)

        self.get_table_columns_values = [
            self.ui.action.text,
            self.ui.wait.isChecked,
            self.ui.tracks.isChecked,
            self.ui.refresh.isChecked
        ]

    def _open_file_dialog(self):
        """Adds the selected file's path to the command line text box.
        """
        file = QtWidgets.QFileDialog.getOpenFileName(self)[0]
        cursor_position = self.ui.action.cursorPosition()
        current_text = self.ui.action.text()
        if not sys.IS_WIN:
            file = shlex.quote(file)
        new_text = current_text[:cursor_position] + file + current_text[cursor_position:]
        self.ui.action.setText(new_text)

    def _reset_ui(self):
        self.ui.action.setText("")
        self.ui.wait.setChecked(False)
        self.ui.refresh.setChecked(False)
        self.ui.albums.setChecked(True)

    def _add_action_to_table(self):
        if not self.ui.action.text():
            return
        row_position = self.ui.table.rowCount()
        self.ui.table.insertRow(row_position)
        for column in range(self.ui.table.columnCount()):
            value = self.get_table_columns_values[column]()
            value = str(value)
            widget = QtWidgets.QTableWidgetItem(value)
            self.ui.table.setItem(row_position, column, widget)
        self._reset_ui()

    def _remove_action_from_table(self):
        current_row = self.ui.table.currentRow()
        if current_row != -1:
            self.ui.table.removeRow(current_row)

    def _move_action_up(self):
        current_row = self.ui.table.currentRow()
        new_row = current_row - 1
        if current_row > 0:
            self._swap_table_rows(current_row, new_row)
            self.ui.table.setCurrentCell(new_row, 0)

    def _move_action_down(self):
        current_row = self.ui.table.currentRow()
        new_row = current_row + 1
        if current_row < self.ui.table.rowCount() - 1:
            self._swap_table_rows(current_row, new_row)
            self.ui.table.setCurrentCell(new_row, 0)

    def _swap_table_rows(self, row1, row2):
        for column in range(self.ui.table.columnCount()):
            item1 = self.ui.table.takeItem(row1, column)
            item2 = self.ui.table.takeItem(row2, column)
            self.ui.table.setItem(row1, column, item2)
            self.ui.table.setItem(row2, column, item1)

    def load(self):
        """Puts the plugin's settings into the actions table.
        """
        settings = zip(*[config.setting[name] for name in OPTIONS])
        for row, values in enumerate(settings):
            self.ui.table.insertRow(row)
            for column in range(self.ui.table.columnCount()):
                widget = QtWidgets.QTableWidgetItem(values[column])
                self.ui.table.setItem(row, column, widget)
        self.ui.cancel.setChecked(config.setting[CANCEL])
        self.ui.max_workers.setValue(config.setting[MAX_WORKERS])

    def save(self):
        """Saves the actions table items in the settings.
        """
        settings = []
        for column in range(self.ui.table.columnCount()):
            settings.append([])
            for row in range(self.ui.table.rowCount()):
                setting = self.ui.table.item(row, column).text()
                settings[column].append(setting)
            config.setting[OPTIONS[column]] = settings[column]
        config.setting[CANCEL] = self.ui.cancel.isChecked()
        config.setting[MAX_WORKERS] = self.ui.max_workers.value()
        action_loader.load_actions()


class ActionsStatus(QtWidgets.QWidget, Ui_ActionsStatus):
    """An icon and a label that displays the number of pending actions.

    The widget is only visible when there are pending actions. This is placed
    in the statusbar to the left of the other icons.
    """

    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        Ui_ActionsStatus.__init__(self)
        self.setupUi(self)
        self.hide()

        # Creates the icon to the right of the label.
        size = QtCore.QSize(16, 16)
        icon = QtGui.QIcon(":/images/16x16/applications-system.png")
        self.label.setPixmap(icon.pixmap(size))

    def update_actions_count(self, count):
        self.actions_count.setText(f"{count}")
        self.setVisible(count > 0)


action_loader = ActionLoader()
action_runner = ActionRunner()
register_album_action(ExecuteAlbumActions())
register_track_action(ExecuteTrackActions())
register_options_page(PostTaggingActionsOptions)



PLUGIN_NAME = 'Disc Numbers'
PLUGIN_AUTHOR = 'Lukas Lalinsky'
PLUGIN_DESCRIPTION = '''Moves disc numbers and subtitles from album titles to separate tags. For example:<br/>
<em>"Aerial (disc 1: A Sea of Honey)"</em>
<ul>
    <li>album = <em>"Aerial"</em></li>
    <li>discnumber = <em>"1"</em></li>
    <li>discsubtitle = <em>"A Sea of Honey"</em></li>
</ul>'''
PLUGIN_VERSION = "0.1"
PLUGIN_API_VERSIONS = ["0.9.0", "0.10", "0.15", "2.0"]

from picard.metadata import register_album_metadata_processor
import re

_discnumber_re = re.compile(r"\s+\(disc (\d+)(?::\s+([^)]+))?\)")


def remove_discnumbers(tagger, metadata, release):
    matches = _discnumber_re.search(metadata["album"])
    if matches:
        metadata["discnumber"] = matches.group(1)
        if matches.group(2):
            metadata["discsubtitle"] = matches.group(2)
        metadata["album"] = _discnumber_re.sub('', metadata["album"])

register_album_metadata_processor(remove_discnumbers)



# -*- coding: utf-8 -*-

# This is the Sort Multivalue Tags plugin for MusicBrainz Picard.
# Copyright (C) 2013 Sophist
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Abbreviate artist-sort"
PLUGIN_AUTHOR = "Sophist"
PLUGIN_DESCRIPTION = '''Abbreviate Artist-Sort and Album-Artist-Sort Tags.
e.g. "Vivaldi, Antonio" becomes "Vivaldi, A."
This is particularly useful for classical albums that can have a long list of artists.
%artistsort% is abbreviated into %_artistsort_abbrev% and
%albumartistsort% is abbreviated into %_albumartistsort_abbrev%.'''
PLUGIN_VERSION = "0.4.1"
PLUGIN_API_VERSIONS = ["1.0", "2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


from picard import log
from picard.metadata import register_track_metadata_processor

# NOTE: This plugin will not work consistently if you have not enabled the 'Standardize Artist Names' option!
# The algorithm for this is complicated because the tags can contain multiple names separated by various characters
# As an example from http://musicbrainz.org/release/6c0cfb20-2606-46c1-9306-ee5e7cb5bfdf
#   Sorted:   Vivaldi, Antonio, Caldara, Antonio; Queyras, Jean-Guihen, Kallweit, Georg, Akademie für Alte Musik Berlin
#   Unsorted: Antonio Vivaldi, Antonio Caldara; Jean-Guihen Queyras, Georg Kallweit, Akademie für Alte Musik Berlin
# As you can see, in unsorted, names are separated by ',' as well as ';' but could be e.g. 'feat:'
# The only known is that in sorted, surname is separated from forename(s) by a ','
# It is further complicated by non-latin (e.g. japanese) script artist names
#   where unsorted is in local script, but sorted can be in latin, and names can be in local locale or general.

# If the names are just reversed, and we shift Unsorted to the right by one character (for the ',' in sorted),
# then the punctuation should start to match up:
#   Sorted:   Vivaldi, Antonio, Caldara, Antonio; Queyras, Jean-Guihen, Kallweit, Georg, Akademie für Alte Musik Berlin
#   Unsorted:  Antonio Vivaldi, Antonio Caldara; Jean-Guihen Queyras, Georg Kallweit, Akademie für Alte Musik Berlin
#                             ^
# Of course if we have non-latin or locale names, alignment could be way off:
#   Sorted: Verdi, Giuseppe, Vivaldi, Antonio
#   Unsorted: Joe Green, Antonio Vivaldi

# In the absence of an array version of the tags (PR pending)
# we need to process the tag and sorted tag together in a special way as follows:
#
# 1. Look for the first ',' in sorted and tentatively set surname to the string up to that point.
#   Case a. Sorted:   Stuff, ...
#           Unsorted: Stuff, ...
#   Case b. Sorted:   Surname, Forename(s)...
#           Unsorted: Forename(s) Surname...
#     Special case: Sorted:   Major, Major...
#                   Unsorted: Major Major...
#   Case c. Sorted:   Stuff; Surname, Forename(s)...
#           Unsorted: Stuff; Forename(s) Surname...
#   Case d. Sorted:   Latin Surname, Latin Forename(s)...
#           Unsorted: Foreign...
#   Case e. Sorted:   Beatles, The...
#           Unsorted: The Beatles...
# 2. Locate surname in unsorted:
#   Case a. unsorted starts with surname - move both to new strings
#   Case b. surname can be found in unsorted - forename(s) are what is before and match beginning of rest
#   Case c. If first word is same in sorted and unsorted, move words that match to new strings, then treat as b.
#   Case d. Try to handle without abbreviating and get to next name which might not be foreign

_abbreviate_tags = [
    ('albumartistsort', 'albumartist', '~albumartistsort_abbrev'),
    ('artistsort', 'artist', '~artistsort_abbrev'),
]
_prefixes = ["A", "The"]
_split = ", "
_abbreviate_cache = {}


def abbreviate_artistsort(tagger, metadata, track, release):

    for sortTag, unsortTag, sortTagNew in _abbreviate_tags:
        if not (sortTag in metadata and unsortTag in metadata):
            continue

        sorts = list(metadata.getall(sortTag))
        unsorts = list(metadata.getall(unsortTag))
        for i in range(0, min(len(sorts), len(unsorts))):
            sort = sorts[i]
            log.debug("%s: Trying to abbreviate '%s'." % (PLUGIN_NAME, sort))
            if sort in _abbreviate_cache:
                log.debug("  Using abbreviation found in cache: '%s'." % (_abbreviate_cache[sort]))
                sorts[i] = _abbreviate_cache[sort]
                continue
            unsort = unsorts[i]
            new_sort = ""
            new_unsort = ""

            while len(sort) > 0 and len(unsort) > 0:

                if not _split in sort:
                    log.debug("  Ending without separator '%s' - moving '%s'." % (_split, sort))
                    new_sort += sort
                    new_unsort += unsort
                    sort = unsort = ""
                    continue

                surname, rest = sort.split(_split, 1)
                if rest == "":
                    log.debug("  Ending with separator '%s' - moving '%s'." % (_split, surname))
                    new_sort += sort
                    new_unsort += unsort
                    sort = unsort = ""
                    continue

                # Move leading whitespace
                new_unsort += unsort[0:len(unsort) - len(unsort.lstrip())]
                unsort = unsort.lstrip()

                # Sorted:   Stuff, ...
                # Unsorted: Stuff, ...
                temp = surname + _split
                l = len(temp)
                if unsort[:l] == temp:
                    log.debug("  No forename - moving '%s'." % (surname))
                    new_sort += temp
                    new_unsort += temp
                    sort = sort[l:]
                    unsort = unsort[l:]
                    continue

                # Sorted:   Stuff; Surname, Forename(s)...
                # Unsorted: Stuff; Forename(s) Surname...
                # Move matching words plus white-space one by one
                if unsort.find(' ' + surname) == -1:
                    while surname.split(None, 1)[0] == unsort.split(None, 1)[0]:
                        x = unsort.split(None, 1)[0]
                        log.debug("  Moving matching word '%s'." % (x))
                        new_sort += x
                        new_unsort += x
                        surname = surname[len(x):]
                        unsort = unsort[len(x):]
                        new_sort += surname[0:len(surname) - len(surname.lstrip())]
                        surname = surname.lstrip()
                        new_unsort += unsort[0:len(unsort) - len(unsort.lstrip())]
                        unsort = unsort.lstrip()

                # If we still can't find surname then we are up a creek...
                pos = unsort.find(' ' + surname)
                if pos == -1:
                    log.debug(
                        _("%s: Track %s: Unable to abbreviate surname '%s' - not matched in unsorted %s: '%s'."),
                        PLUGIN_NAME,
                        metadata['tracknumber'],
                        surname,
                        unsortTag,
                        unsort[i],
                    )
                    log.warning("  Could not match surname '%s' in remaining unsorted: %s" % (surname, unsort))
                    break

                # Sorted:   Surname, Forename(s)...
                # Unsorted: Forename(s) Surname...
                forename = unsort[:pos]
                if rest[:len(forename)] != forename:
                    log.debug(
                        _("%s: Track %s: Unable to abbreviate surname (%s) - forename (%s) not matched in unsorted %s: '%s'."),
                        PLUGIN_NAME,
                        metadata['tracknumber'],
                        surname,
                        forename,
                        unsortTag,
                        unsort[i],
                    )
                    log.warning("  Could not match forename (%s) for surname (%s) in remaining unsorted (%s):" % (forename, surname, unsort))
                    break

                inits = ' '.join([x[0] + '.' for x in forename.split()])

                # Sorted:   Beatles, The...
                # Unsorted: The Beatles...
                if forename in _prefixes:
                    inits = forename

                new_sort += surname + _split + inits
                sort = rest[len(forename):]
                new_sort += sort[0:len(sort) - len(sort[1:].lstrip())]
                sort = sort[1:].lstrip()
                new_unsort += forename
                unsort = unsort[len(forename):]
                new_unsort += unsort[0:len(unsort) - len(unsort.lstrip())]
                unsort = unsort.lstrip()
                new_unsort += surname
                unsort = unsort[len(surname):]
                new_unsort += unsort[0:len(unsort) - len(unsort[1:].lstrip())]
                unsort = unsort[1:].lstrip()

                if forename != inits:
                    log.debug(
                        _("%s: Abbreviated surname (%s, %s) to (%s, %s) in '%s'."),
                        PLUGIN_NAME,
                        surname,
                        forename,
                        surname,
                        inits,
                        sortTag,
                    )
                    log.debug("Abbreviated (%s, %s) to (%s, %s)." % (surname, forename, surname, inits))
            else:  # while loop ended without a break i.e. no errors
                if unsorts[i] != new_unsort:
                    log.error(
                        _("%s: Track %s: Logic error - mangled %s from '%s' to '%s'."),
                        PLUGIN_NAME,
                        metadata['tracknumber'],
                        unsortTag,
                        unsorts[i],
                        new_unsort,
                    )
                    log.warning("Error: Unsorted text for %s has changed from '%s' to '%s'!" % (unsortTag, unsorts[i], new_unsort))
                _abbreviate_cache[sorts[i]] = new_sort
                log.debug("  Abbreviated and cached (%s) as (%s)." % (sorts[i], new_sort))
                if sorts[i] != new_sort:
                    log.debug(_("%s: Abbreviated tag '%s' to '%s'."),
                              PLUGIN_NAME,
                              sorts[i],
                              new_sort,
                              )
                    sorts[i] = new_sort
        metadata[sortTagNew] = sorts

register_track_metadata_processor(abbreviate_artistsort)



# -*- coding: utf-8 -*-
# Acousticbrainz Tonal/Rhythm plugin for Picard
# Copyright (C) 2015  Sophist
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

PLUGIN_NAME = 'AcousticBrainz Tonal-Rhythm'
PLUGIN_AUTHOR = 'Sophist, Sambhav Kothari'
PLUGIN_DESCRIPTION = '''Add's the following tags:
<ul>
<li>Key (in ID3v2.3 format)</li>
<li>Beats Per Minute (BPM)</li>
</ul>
from the AcousticBrainz database.<br/><br/>
<em>This plugin is deprecated, please consider using the AcousticBrainz Tags
plugin instead.</em>
'''
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"
PLUGIN_VERSION = '1.1.6'
PLUGIN_API_VERSIONS = ["2.0"]  # Requires support for TKEY which is in 1.4

from json import JSONDecodeError

from picard import log
from picard.metadata import register_track_metadata_processor
from functools import partial
from picard.webservice import ratecontrol
from picard.util import load_json

ACOUSTICBRAINZ_HOST = "acousticbrainz.org"
ACOUSTICBRAINZ_PORT = 80

ratecontrol.set_minimum_delay((ACOUSTICBRAINZ_HOST, ACOUSTICBRAINZ_PORT), 1000)


class AcousticBrainz_Key:

    def get_data(self, album, track_metadata, track_node, release_node):
        if "musicbrainz_recordingid" not in track_metadata:
            log.error("%s: Error parsing response. No MusicBrainz recording id found.",
                      PLUGIN_NAME)
            return
        recordingId = track_metadata['musicbrainz_recordingid']
        if recordingId:
            log.debug("%s: Add AcousticBrainz request for %s (%s)",
                      PLUGIN_NAME, track_metadata['title'], recordingId)
            self.album_add_request(album)
            path = "/%s/low-level" % recordingId
            return album.tagger.webservice.get(
                        ACOUSTICBRAINZ_HOST,
                        ACOUSTICBRAINZ_PORT,
                        path,
                        partial(self.process_data, album, track_metadata),
                        priority=True,
                        important=False,
                        parse_response_type=None)

    def process_data(self, album, track_metadata, response, reply, error):
        if error:
            log.error("%s: Network error retrieving acousticBrainz data for recordingId %s",
                      PLUGIN_NAME, track_metadata['musicbrainz_recordingid'])
            self.album_remove_request(album)
            return
        try:
            data = load_json(response)
        except JSONDecodeError:
            log.error("%s: Network error retrieving AcousticBrainz data for recordingId %s",
                      PLUGIN_NAME, track_metadata['musicbrainz_recordingid'])
            self.album_remove_request(album)
            return
        if "tonal" in data:
            if "key_key" in data["tonal"]:
                key = data["tonal"]["key_key"]
                if "key_scale" in data["tonal"]:
                    scale = data["tonal"]["key_scale"]
                    if scale == "minor":
                        key += "m"
                track_metadata["key"] = key
                log.debug("%s: Track '%s' is in key %s", PLUGIN_NAME, track_metadata["title"], key)
        if "rhythm" in data:
            if "bpm" in data["rhythm"]:
                bpm = int(data["rhythm"]["bpm"] + 0.5)
                track_metadata["bpm"] = bpm
                log.debug("%s: Track '%s' has %s bpm", PLUGIN_NAME, track_metadata["title"], bpm)
        self.album_remove_request(album)

    def album_add_request(self, album):
        album._requests += 1

    def album_remove_request(self, album):
        album._requests -= 1
        if album._requests == 0:
            album._finalize_loading(None)


register_track_metadata_processor(AcousticBrainz_Key().get_data)



# -*- coding: utf-8 -*-

# Changelog:
# [2015-09-24] Initial version with support for Ogg Vorbis, FLAC, WAV and MP3, tested MP3 and FLAC
# [2017-11-21] Amended to Python3 & Qt5
# [2017-11-21] removed unicode, replaced str with string_ and untrusted input on check_call addressed

PLUGIN_NAME = "Moodbars"
PLUGIN_AUTHOR = "Len Joubert, Sambhav Kothari"
PLUGIN_DESCRIPTION = """Calculate Moodbars for selected files and albums.<br /><br />
According to <a href="http://en.wikipedia.org/wiki/Moodbar">WikiPedia</a>
a "Moodbar is a computer visualization used for navigating within a piece of music or any other recording on a digital audio track.
This is done with a commonly horizontal bar that is divided into vertical stripes.
Each stripe has a colour combination showing the "mood" within a short part of the audio track."<br /><br />
To use this plugin you will need to download special executables to create the moodbars -
at the time of writing, executables are only available for various Linux distributions
(see the <a href="http://userbase.kde.org/Amarok/Manual/Various/Moodbar">Amarok Moodbar page<a> for details).
"""
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"
PLUGIN_VERSION = "2.3.3"
PLUGIN_API_VERSIONS = ["2.0"]
# PLUGIN_INCOMPATIBLE_PLATFORMS = [
#    'win32', 'cygwyn', 'darwin', 'os2', 'os2emx', 'riscos', 'atheos']

import os.path
from functools import partial
from collections import defaultdict
from subprocess import check_call
from picard.album import Album, NatAlbum
from picard.track import Track
from picard.file import File
from picard.util import encode_filename, decode_filename, thread
from picard.ui.options import register_options_page, OptionsPage
from picard.config import TextOption
from picard.ui.itemviews import (BaseAction, register_file_action,
                                 register_album_action)
from picard.plugins.moodbars.ui_options_moodbar import Ui_MoodbarOptionsPage

# Path to various moodbar tools. There must be a tool for every supported
# audio file format.
MOODBAR_COMMANDS = {
    "Ogg Vorbis": ("moodbar_vorbis_command", "moodbar_vorbis_options"),
    "MPEG-1 Audio": ("moodbar_mp3_command", "moodbar_mp3_options"),
    "FLAC": ("moodbar_flac_command", "moodbar_flac_options"),
    "WavPack": ("moodbar_wav_command", "moodbar_wav_options"),
}


def generate_moodbar_for_files(files, format, tagger):
    """Generate the moodfiles for a list of files in album mode."""
    # python3 list for subprocess
    command_to_execute = []
    file_list = [files]
    for mood_file in file_list:
        new_filename = os.path.join(os.path.dirname(
            mood_file), '.' + os.path.splitext(os.path.basename(mood_file))[0] + '.mood')
        # file format to make it compaitble with Amarok and hidden in linux
        file_list_mood = ['%s' % new_filename]
        file_list_music = ['%s' % file_list[0]]

    if format in MOODBAR_COMMANDS \
            and tagger.config.setting[MOODBAR_COMMANDS[format][0]]:
        command = tagger.config.setting[MOODBAR_COMMANDS[format][0]]
        options = tagger.config.setting[
            MOODBAR_COMMANDS[format][1]].split(' ')
        # tagger.log.debug('My debug file_list_mood >>>  %s' %
        #    (file_list_mood))
        #tagger.log.debug('My debug file_list >>>  %s' % (file_list_music))
        #tagger.log.debug('My debug command >>>  %s' % (command))
        #tagger.log.debug('My debug options >>>  %s' % (options))
        # tagger.log.debug(
        #    '%s %s %s %s' % (command, decode_filename(' '.join(file_list)), ' '.join(options), decode_filename(' '.join(file_list_mood))))
        # command args order corrected for new moodbar
        command_to_execute.append(command)
        command_to_execute = command_to_execute + options
        # need to decode the file string and appand to the list
        command_to_execute.append(file_list_mood[0])
        command_to_execute = command_to_execute + file_list_music
        tagger.log.debug('My debug command to execute >>>  %s' %
                         (command_to_execute))
        check_call(command_to_execute, shell=False)
        # check_call([command] + options + file_list_mood + file_list)
    else:
        raise Exception('Moodbar: Unsupported format %s' % (format))


class MoodBar(BaseAction):
    NAME = N_("Generate Moodbar &file...")

    def _add_file_to_queue(self, file):
        thread.run_task(
            partial(self._generate_moodbar, file),
            partial(self._moodbar_callback, file))

    def callback(self, objs):
        for obj in objs:
            if isinstance(obj, Track):
                for file_ in obj.linked_files:
                    self._add_file_to_queue(file_)
            elif isinstance(obj, File):
                self._add_file_to_queue(obj)

    def _generate_moodbar(self, file):
        self.tagger.window.set_statusbar_message(
            N_('Calculating moodbar for "%(filename)s"...'),
            {'filename': file.filename}
        )
        generate_moodbar_for_files(file.filename, file.NAME, self.tagger)

    def _moodbar_callback(self, file, result=None, error=None):
        if not error:
            self.tagger.window.set_statusbar_message(
                N_('Moodbar for "%(filename)s" successfully generated.'),
                {'filename': file.filename}
            )
        else:
            self.tagger.window.set_statusbar_message(
                N_('Could not generate moodbar for "%(filename)s".'),
                {'filename': file.filename}
            )


class MoodbarOptionsPage(OptionsPage):

    NAME = "Moodbars"
    TITLE = "Moodbars"
    PARENT = "plugins"

    options = [
        TextOption("setting", "moodbar_vorbis_command", "moodbar"),
        TextOption("setting", "moodbar_vorbis_options", "-o"),
        TextOption("setting", "moodbar_mp3_command", "moodbar"),
        TextOption("setting", "moodbar_mp3_options", "-o"),
        TextOption("setting", "moodbar_flac_command", "moodbar"),
        TextOption("setting", "moodbar_flac_options", "-o"),
        TextOption("setting", "moodbar_wav_command", "moodbar"),
        TextOption("setting", "moodbar_wav_options", "-o")
    ]

    def __init__(self, parent=None):
        super(MoodbarOptionsPage, self).__init__(parent)
        self.ui = Ui_MoodbarOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        self.ui.vorbis_command.setText(
            self.config.setting["moodbar_vorbis_command"])
        self.ui.mp3_command.setText(
            self.config.setting["moodbar_mp3_command"])
        self.ui.flac_command.setText(
            self.config.setting["moodbar_flac_command"])
        self.ui.wav_command.setText(
            self.config.setting["moodbar_wav_command"])

    def save(self):
        self.config.setting["moodbar_vorbis_command"] = self.ui.vorbis_command.text()
        self.config.setting["moodbar_mp3_command"] = self.ui.mp3_command.text()
        self.config.setting["moodbar_flac_command"] = self.ui.flac_command.text()
        self.config.setting["moodbar_wav_command"] = self.ui.wav_command.text()

register_file_action(MoodBar())
register_options_page(MoodbarOptionsPage)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014, 2017, 2021, 2023-2024 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

PLUGIN_NAME = 'Tagger script functions $is_lossless() and $is_lossy()'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = 'Tagger script functions to detect if a file is lossless or lossy'
PLUGIN_VERSION = "0.3"
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2", "2.3", "2.4", "2.5", "2.6", "2.7", "2.8", "2.9", "2.10", "2.11"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


LOSSLESS_EXTENSIONS = {'flac', 'oggflac', 'ape', 'ofr', 'tak', 'wv', 'tta',
                       'aiff', 'aif', 'wav'}

from picard.script import register_script_function


def is_lossless(parser):
    """
    Returns true, if the file processed is a lossless audio format.
    Note: This depends mainly on the file extension and does not look inside the
    file to detect the codec.
    """
    if parser.context['~extension'] in LOSSLESS_EXTENSIONS:
        return "1"
    elif parser.context['~extension'] == 'm4a':
        # Mutagen < 1.26 fails to detect the bitrate for Apple Lossless Audio Codec.
        if not parser.context['~bitrate']:
            return "1"
        else:
            try:
                bitrate = float(parser.context['~bitrate'])
                return "1" if bitrate > 1000 else ""
            except (TypeError, ValueError):
                return ""
    else:
        return ""


def is_lossy(parser):
    """Returns true, if the file processed is a lossy audio format."""
    return "" if is_lossless(parser) == "1" else "1"


register_script_function(is_lossless)
register_script_function(is_lossy)



# -*- coding: utf-8 -*-

# Changelog:
# [2015-09-15] Initial version
# [2017-11-24] Qt5, Python3 for Picard-plugins branch 2
# [2020-12-25] Move access to config.settings outside of thread
# Dependancies:
# aubio, numpy
#

PLUGIN_NAME = "BPM Analyzer"
PLUGIN_AUTHOR = "Len Joubert, Sambhav Kothari, Philipp Wolfer"
PLUGIN_DESCRIPTION = """Calculate BPM for selected files and albums. Linux only version with dependancy on Aubio and Numpy"""
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"
PLUGIN_VERSION = "1.5.1"
PLUGIN_API_VERSIONS = ["2.0"]
# PLUGIN_INCOMPATIBLE_PLATFORMS = [
#    'win32', 'cygwyn', 'darwin', 'os2', 'os2emx', 'riscos', 'atheos']

from aubio import source, tempo
from functools import partial
from numpy import median, diff

from picard.config import config, IntOption
from picard.file import File
from picard.plugins.bpm.ui_options_bpm import Ui_BPMOptionsPage
from picard.track import Track
from picard.ui.itemviews import BaseAction, register_file_action
from picard.ui.options import register_options_page, OptionsPage
from picard.util import thread


bpm_slider_settings = {
    1: (44100, 1024, 512),
    2: (8000, 512, 128),
    3: (4000, 128, 64),
}


class FileBPM(BaseAction):
    NAME = N_("Calculate BPM...")

    def __init__(self):
        super().__init__()
        self._close = False
        self.tagger.aboutToQuit.connect(self._cleanup)

    def _cleanup(self):
        self._close = True

    def _add_file_to_queue(self, file):
        settings = bpm_slider_settings[config.setting["bpm_slider_parameter"]]
        thread.run_task(
            partial(self._calculate_bpm, file, settings),
            partial(self._calculate_bpm_callback, file))

    def callback(self, objs):
        for obj in objs:
            if isinstance(obj, Track):
                for file_ in obj.linked_files:
                    self._add_file_to_queue(file_)
            elif isinstance(obj, File):
                self._add_file_to_queue(obj)

    def _calculate_bpm(self, file, settings):
        self.tagger.window.set_statusbar_message(
            N_('Calculating BPM for "%(filename)s"...'),
            {'filename': file.filename}
        )
        calculated_bpm = self._get_file_bpm(file.filename, settings)
        # self.tagger.log.debug('%s' % (calculated_bpm))
        if self._close:
            return
        file.metadata["bpm"] = str(round(calculated_bpm, 1))
        file.update()

    def _calculate_bpm_callback(self, file, result=None, error=None):
        if not error:
            self.tagger.window.set_statusbar_message(
                N_('BPM for "%(filename)s" successfully calculated.'),
                {'filename': file.filename}
            )
        else:
            self.tagger.window.set_statusbar_message(
                N_('Could not calculate BPM for "%(filename)s".'),
                {'filename': file.filename}
            )

    def _get_file_bpm(self, path, settings):
        """ Calculate the beats per minute (bpm) of a given file.
            path: path to the file
            buf_size    length of FFT
            hop_size    number of frames between two consecutive runs
            samplerate  sampling rate of the signal to analyze
        """

        samplerate, buf_size, hop_size = settings
        mediasource = source(path, samplerate, hop_size)
        samplerate = mediasource.samplerate
        beattracking = tempo("specdiff", buf_size, hop_size, samplerate)
        # List of beats, in samples
        beats = []
        # Total number of frames read
        total_frames = 0

        while True:
            if self._close:
                return
            samples, read = mediasource()
            is_beat = beattracking(samples)
            if is_beat:
                this_beat = beattracking.get_last_s()
                beats.append(this_beat)
            total_frames += read
            if read < hop_size:
                break

        # Convert to periods and to bpm
        bpms = 60. / diff(beats)
        return median(bpms)


class BPMOptionsPage(OptionsPage):

    NAME = "bpm"
    TITLE = "BPM"
    PARENT = "plugins"
    ACTIVE = True

    options = [
        IntOption("setting", "bpm_slider_parameter", 1)
    ]

    def __init__(self, parent=None):
        super(BPMOptionsPage, self).__init__(parent)
        self.ui = Ui_BPMOptionsPage()
        self.ui.setupUi(self)
        self.ui.slider_parameter.valueChanged.connect(self.update_parameters)
        self.update_parameters()

    def load(self):
        cfg = self.config.setting
        self.ui.slider_parameter.setValue(cfg["bpm_slider_parameter"])

    def save(self):
        cfg = self.config.setting
        cfg["bpm_slider_parameter"] = self.ui.slider_parameter.value()

    def update_parameters(self):
        val = self.ui.slider_parameter.value()
        samplerate, buf_size, hop_size = [str(v) for v in
                                          bpm_slider_settings[val]]
        self.ui.samplerate_value.setText(samplerate)
        self.ui.win_s_value.setText(buf_size)
        self.ui.hop_s_value.setText(hop_size)


register_file_action(FileBPM())
register_options_page(BPMOptionsPage)



# -*- coding: utf-8 -*-

# Copyright © 2015 Samir Benmendil <me@rmz.io>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See http://www.wtfpl.net/ for more details.

PLUGIN_NAME = 'Soundtrack'
PLUGIN_AUTHOR = 'Samir Benmendil'
PLUGIN_LICENSE = 'WTFPL'
PLUGIN_LICENSE_URL = 'http://www.wtfpl.net/'
PLUGIN_DESCRIPTION = '''Sets the albumartist to "Soundtrack" if releasetype is a soundtrack.'''
PLUGIN_VERSION = "0.2"
PLUGIN_API_VERSIONS = ["1.0", "2.0"]

from picard.metadata import register_album_metadata_processor


def soundtrack(tagger, metadata, release):
    if "soundtrack" in metadata["releasetype"]:
        metadata["albumartist"] = "Soundtrack"
        metadata["albumartistsort"] = "Soundtrack"

register_album_metadata_processor(soundtrack)



# -*- coding: utf-8 -*-

PLUGIN_NAME = 'Last.fm'
PLUGIN_AUTHOR = 'Lukáš Lalinský, Philipp Wolfer'
PLUGIN_DESCRIPTION = 'Use tags from Last.fm as genre.'
PLUGIN_VERSION = "0.10.1"
PLUGIN_API_VERSIONS = ["2.0"]

import re
from functools import partial
from PyQt5 import QtCore
from picard import config, log
from picard.config import BoolOption, IntOption, TextOption
from picard.metadata import register_track_metadata_processor
from picard.plugins.lastfm.ui_options_lastfm import Ui_LastfmOptionsPage
from picard.ui.options import register_options_page, OptionsPage
from picard.util import build_qurl
from picard.webservice import ratecontrol

LASTFM_HOST = 'ws.audioscrobbler.com'
LASTFM_PORT = 80
LASTFM_PATH = '/2.0/'
LASTFM_API_KEY = '0a210a4a6741f2ec8f27a791b9d5d971'

# From https://www.last.fm/api/tos, 2018-09-04
# 4.4 […] You will not make more than 5 requests per originating IP address per
# second, averaged over a 5 minute period, without prior written consent. […]
ratecontrol.set_minimum_delay((LASTFM_HOST, LASTFM_PORT), 200)

# Cache for Tags to avoid re-requesting tags within same Picard session
_cache = {}

# Keeps track of requests for tags made to webservice API but not yet returned
# (to avoid re-requesting the same URIs)
_pending_requests = {}

# TODO: move this to an options page
TRANSLATE_TAGS = {
    "hip hop": "Hip-Hop",
    "synth-pop": "Synthpop",
    "electronica": "Electronic",
}
TITLE_CASE = True


def parse_ignored_tags(ignore_tags_setting):
    ignore_tags = []
    for tag in ignore_tags_setting.lower().split(','):
        tag = tag.strip()
        if tag.startswith('/') and tag.endswith('/'):
            try:
                tag = re.compile(tag[1:-1])
            except re.error:
                log.error(
                    'Error parsing ignored tag "%s"', tag, exc_info=True)
        ignore_tags.append(tag)
    return ignore_tags


def matches_ignored(ignore_tags, tag):
    tag = tag.lower().strip()
    for pattern in ignore_tags:
        if hasattr(pattern, 'match'):
            match = pattern.match(tag)
        else:
            match = pattern == tag
        if match:
            return True
    return False


def _tags_finalize(album, metadata, tags, next_):
    if next_:
        next_(tags)
    else:
        tags = list(set(tags))
        if tags:
            join_tags = config.setting["lastfm_join_tags"]
            if join_tags:
                tags = join_tags.join(tags)
            metadata["genre"] = tags


def _tags_downloaded(album, metadata, min_usage, ignore, next_, current, data,
                     reply, error):
    if error:
        album._requests -= 1
        album._finalize_loading(None)
        return

    try:
        try:
            intags = data.lfm[0].toptags[0].tag
        except AttributeError:
            intags = []
        tags = []
        for tag in intags:
            name = tag.name[0].text.strip()
            try:
                count = int(tag.count[0].text.strip())
            except ValueError:
                count = 0
            if count < min_usage:
                break
            try:
                name = TRANSLATE_TAGS[name]
            except KeyError:
                pass
            if not matches_ignored(ignore, name):
                tags.append(name.title())
        url = reply.url().toString()
        _cache[url] = tags
        _tags_finalize(album, metadata, current + tags, next_)

        # Process any pending requests for the same URL
        if url in _pending_requests:
            pending = _pending_requests[url]
            del _pending_requests[url]
            for delayed_call in pending:
                delayed_call()

    except Exception:
        log.error('Problem processing download tags', exc_info=True)
    finally:
        album._requests -= 1
        album._finalize_loading(None)


def get_tags(album, metadata, queryargs, min_usage, ignore, next_, current):
    """Get tags from an URL."""
    url = build_qurl(
        LASTFM_HOST, LASTFM_PORT, LASTFM_PATH, queryargs).toString()
    if url in _cache:
        _tags_finalize(album, metadata, current + _cache[url], next_)
    else:
        # If we have already sent a request for this URL, delay this call
        if url in _pending_requests:
            _pending_requests[url].append(
                partial(get_tags, album, metadata, queryargs, min_usage,
                        ignore, next_, current))
        else:
            _pending_requests[url] = []
            album._requests += 1
            album.tagger.webservice.get(
                LASTFM_HOST, LASTFM_PORT, LASTFM_PATH,
                partial(_tags_downloaded, album, metadata, min_usage, ignore,
                        next_, current),
                queryargs=queryargs, parse_response_type='xml',
                priority=True, important=True)


def encode_str(s):
    s = QtCore.QUrl.toPercentEncoding(s)
    return bytes(s).decode()


def get_queryargs(queryargs):
    queryargs = {k: encode_str(v) for (k, v) in queryargs.items()}
    queryargs['api_key'] = LASTFM_API_KEY
    return queryargs


def get_track_tags(album, metadata, artist, track, min_usage,
                   ignore, next_, current):
    """Get track top tags."""
    queryargs = get_queryargs({
        'method': 'Track.getTopTags',
        'artist': artist,
        'track': track,
    })
    get_tags(album, metadata, queryargs, min_usage, ignore, next_, current)


def get_artist_tags(album, metadata, artist, min_usage,
                    ignore, next_, current):
    """Get artist top tags."""
    queryargs = get_queryargs({
        'method': 'Artist.getTopTags',
        'artist': artist,
    })
    get_tags(album, metadata, queryargs, min_usage, ignore, next_, current)


def process_track(album, metadata, track, release):
    use_track_tags = config.setting["lastfm_use_track_tags"]
    use_artist_tags = config.setting["lastfm_use_artist_tags"]
    min_tag_usage = config.setting["lastfm_min_tag_usage"]
    ignore_tags = parse_ignored_tags(config.setting["lastfm_ignore_tags"])
    if use_track_tags or use_artist_tags:
        artist = metadata["artist"]
        title = metadata["title"]
        if artist:
            if use_artist_tags:
                get_artist_tags_func = partial(get_artist_tags, album,
                                               metadata, artist, min_tag_usage,
                                               ignore_tags, None)
            else:
                get_artist_tags_func = None
            if title and use_track_tags:
                get_track_tags(album, metadata, artist, title, min_tag_usage,
                               ignore_tags, get_artist_tags_func, [])
            elif get_artist_tags_func:
                get_artist_tags_func([])


class LastfmOptionsPage(OptionsPage):

    NAME = "lastfm"
    TITLE = "Last.fm"
    PARENT = "plugins"

    options = [
        BoolOption("setting", "lastfm_use_track_tags", False),
        BoolOption("setting", "lastfm_use_artist_tags", False),
        IntOption("setting", "lastfm_min_tag_usage", 90),
        TextOption("setting", "lastfm_ignore_tags",
                   "seen live, favorites, /\\d+ of \\d+ stars/"),
        TextOption("setting", "lastfm_join_tags", ""),
    ]

    def __init__(self, parent=None):
        super(LastfmOptionsPage, self).__init__(parent)
        self.ui = Ui_LastfmOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        setting = config.setting
        self.ui.use_track_tags.setChecked(setting["lastfm_use_track_tags"])
        self.ui.use_artist_tags.setChecked(setting["lastfm_use_artist_tags"])
        self.ui.min_tag_usage.setValue(setting["lastfm_min_tag_usage"])
        self.ui.ignore_tags.setText(setting["lastfm_ignore_tags"])
        self.ui.join_tags.setEditText(setting["lastfm_join_tags"])

    def save(self):
        global _cache
        setting = config.setting
        if setting["lastfm_min_tag_usage"] != self.ui.min_tag_usage.value() \
           or setting["lastfm_ignore_tags"] != str(self.ui.ignore_tags.text()):
            _cache = {}
        setting["lastfm_use_track_tags"] = self.ui.use_track_tags.isChecked()
        setting["lastfm_use_artist_tags"] = self.ui.use_artist_tags.isChecked()
        setting["lastfm_min_tag_usage"] = self.ui.min_tag_usage.value()
        setting["lastfm_ignore_tags"] = str(self.ui.ignore_tags.text())
        setting["lastfm_join_tags"] = str(self.ui.join_tags.currentText())


register_track_metadata_processor(process_track)
register_options_page(LastfmOptionsPage)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/format_performer_tags/ui_options_format_performer_tags.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FormatPerformerTagsOptionsPage(object):
    def setupUi(self, FormatPerformerTagsOptionsPage):
        FormatPerformerTagsOptionsPage.setObjectName("FormatPerformerTagsOptionsPage")
        FormatPerformerTagsOptionsPage.resize(561, 802)
        FormatPerformerTagsOptionsPage.setMinimumSize(QtCore.QSize(360, 0))
        self.verticalLayout = QtWidgets.QVBoxLayout(FormatPerformerTagsOptionsPage)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(FormatPerformerTagsOptionsPage)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, -13, 529, 848))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gb_description = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_description.setObjectName("gb_description")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.gb_description)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.format_description = QtWidgets.QLabel(self.gb_description)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.format_description.sizePolicy().hasHeightForWidth())
        self.format_description.setSizePolicy(sizePolicy)
        self.format_description.setWordWrap(True)
        self.format_description.setObjectName("format_description")
        self.verticalLayout_3.addWidget(self.format_description)
        self.verticalLayout_2.addWidget(self.gb_description)
        self.gb_word_groups = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_word_groups.setObjectName("gb_word_groups")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.gb_word_groups)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.group_additonal = QtWidgets.QGroupBox(self.gb_word_groups)
        self.group_additonal.setObjectName("group_additonal")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.group_additonal)
        self.horizontalLayout.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.additional_rb_1 = QtWidgets.QRadioButton(self.group_additonal)
        self.additional_rb_1.setMaximumSize(QtCore.QSize(40, 16777215))
        self.additional_rb_1.setObjectName("additional_rb_1")
        self.horizontalLayout.addWidget(self.additional_rb_1)
        self.additional_rb_2 = QtWidgets.QRadioButton(self.group_additonal)
        self.additional_rb_2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.additional_rb_2.setObjectName("additional_rb_2")
        self.horizontalLayout.addWidget(self.additional_rb_2)
        self.additional_rb_3 = QtWidgets.QRadioButton(self.group_additonal)
        self.additional_rb_3.setMaximumSize(QtCore.QSize(40, 16777215))
        self.additional_rb_3.setObjectName("additional_rb_3")
        self.horizontalLayout.addWidget(self.additional_rb_3)
        self.additional_rb_4 = QtWidgets.QRadioButton(self.group_additonal)
        self.additional_rb_4.setMaximumSize(QtCore.QSize(40, 16777215))
        self.additional_rb_4.setObjectName("additional_rb_4")
        self.horizontalLayout.addWidget(self.additional_rb_4)
        self.verticalLayout_5.addWidget(self.group_additonal)
        self.group_guest = QtWidgets.QGroupBox(self.gb_word_groups)
        self.group_guest.setObjectName("group_guest")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.group_guest)
        self.horizontalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.guest_rb_1 = QtWidgets.QRadioButton(self.group_guest)
        self.guest_rb_1.setMaximumSize(QtCore.QSize(40, 16777215))
        self.guest_rb_1.setObjectName("guest_rb_1")
        self.horizontalLayout_2.addWidget(self.guest_rb_1)
        self.guest_rb_2 = QtWidgets.QRadioButton(self.group_guest)
        self.guest_rb_2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.guest_rb_2.setObjectName("guest_rb_2")
        self.horizontalLayout_2.addWidget(self.guest_rb_2)
        self.guest_rb_3 = QtWidgets.QRadioButton(self.group_guest)
        self.guest_rb_3.setMaximumSize(QtCore.QSize(40, 16777215))
        self.guest_rb_3.setObjectName("guest_rb_3")
        self.horizontalLayout_2.addWidget(self.guest_rb_3)
        self.guest_rb_4 = QtWidgets.QRadioButton(self.group_guest)
        self.guest_rb_4.setMaximumSize(QtCore.QSize(40, 16777215))
        self.guest_rb_4.setObjectName("guest_rb_4")
        self.horizontalLayout_2.addWidget(self.guest_rb_4)
        self.verticalLayout_5.addWidget(self.group_guest)
        self.group_solo = QtWidgets.QGroupBox(self.gb_word_groups)
        self.group_solo.setObjectName("group_solo")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.group_solo)
        self.horizontalLayout_3.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.solo_rb_1 = QtWidgets.QRadioButton(self.group_solo)
        self.solo_rb_1.setMaximumSize(QtCore.QSize(40, 16777215))
        self.solo_rb_1.setObjectName("solo_rb_1")
        self.horizontalLayout_3.addWidget(self.solo_rb_1)
        self.solo_rb_2 = QtWidgets.QRadioButton(self.group_solo)
        self.solo_rb_2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.solo_rb_2.setObjectName("solo_rb_2")
        self.horizontalLayout_3.addWidget(self.solo_rb_2)
        self.solo_rb_3 = QtWidgets.QRadioButton(self.group_solo)
        self.solo_rb_3.setMaximumSize(QtCore.QSize(40, 16777215))
        self.solo_rb_3.setObjectName("solo_rb_3")
        self.horizontalLayout_3.addWidget(self.solo_rb_3)
        self.solo_rb_4 = QtWidgets.QRadioButton(self.group_solo)
        self.solo_rb_4.setMaximumSize(QtCore.QSize(40, 16777215))
        self.solo_rb_4.setObjectName("solo_rb_4")
        self.horizontalLayout_3.addWidget(self.solo_rb_4)
        self.verticalLayout_5.addWidget(self.group_solo)
        self.group_vocals = QtWidgets.QGroupBox(self.gb_word_groups)
        self.group_vocals.setObjectName("group_vocals")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.group_vocals)
        self.horizontalLayout_4.setContentsMargins(1, 1, 1, 1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.vocals_rb_1 = QtWidgets.QRadioButton(self.group_vocals)
        self.vocals_rb_1.setMinimumSize(QtCore.QSize(40, 20))
        self.vocals_rb_1.setMaximumSize(QtCore.QSize(40, 16777215))
        self.vocals_rb_1.setObjectName("vocals_rb_1")
        self.horizontalLayout_4.addWidget(self.vocals_rb_1)
        self.vocals_rb_2 = QtWidgets.QRadioButton(self.group_vocals)
        self.vocals_rb_2.setMaximumSize(QtCore.QSize(40, 16777215))
        self.vocals_rb_2.setObjectName("vocals_rb_2")
        self.horizontalLayout_4.addWidget(self.vocals_rb_2)
        self.vocals_rb_3 = QtWidgets.QRadioButton(self.group_vocals)
        self.vocals_rb_3.setMaximumSize(QtCore.QSize(40, 16777215))
        self.vocals_rb_3.setObjectName("vocals_rb_3")
        self.horizontalLayout_4.addWidget(self.vocals_rb_3)
        self.vocals_rb_4 = QtWidgets.QRadioButton(self.group_vocals)
        self.vocals_rb_4.setMaximumSize(QtCore.QSize(40, 16777215))
        self.vocals_rb_4.setObjectName("vocals_rb_4")
        self.horizontalLayout_4.addWidget(self.vocals_rb_4)
        self.verticalLayout_5.addWidget(self.group_vocals)
        self.verticalLayout_2.addWidget(self.gb_word_groups)
        self.gb_group_settings = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_group_settings.setObjectName("gb_group_settings")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.gb_group_settings)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setVerticalSpacing(1)
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QtWidgets.QLabel(self.gb_group_settings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.label_3 = QtWidgets.QLabel(self.gb_group_settings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 3, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.format_group_1_start_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_1_start_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_1_start_char.setObjectName("format_group_1_start_char")
        self.gridLayout.addWidget(self.format_group_1_start_char, 1, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.label = QtWidgets.QLabel(self.gb_group_settings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.label_4 = QtWidgets.QLabel(self.gb_group_settings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 4, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.format_group_1_sep_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_1_sep_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_1_sep_char.setObjectName("format_group_1_sep_char")
        self.gridLayout.addWidget(self.format_group_1_sep_char, 1, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_1_end_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_1_end_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_1_end_char.setObjectName("format_group_1_end_char")
        self.gridLayout.addWidget(self.format_group_1_end_char, 1, 3, 1, 1, QtCore.Qt.AlignHCenter)
        self.label_5 = QtWidgets.QLabel(self.gb_group_settings)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 0, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.label_6 = QtWidgets.QLabel(self.gb_group_settings)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.label_7 = QtWidgets.QLabel(self.gb_group_settings)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 0, 3, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_2_start_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_2_start_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_2_start_char.setObjectName("format_group_2_start_char")
        self.gridLayout.addWidget(self.format_group_2_start_char, 2, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_3_start_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_3_start_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_3_start_char.setObjectName("format_group_3_start_char")
        self.gridLayout.addWidget(self.format_group_3_start_char, 3, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_4_start_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_4_start_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_4_start_char.setObjectName("format_group_4_start_char")
        self.gridLayout.addWidget(self.format_group_4_start_char, 4, 1, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_2_sep_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_2_sep_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_2_sep_char.setObjectName("format_group_2_sep_char")
        self.gridLayout.addWidget(self.format_group_2_sep_char, 2, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_3_sep_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_3_sep_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_3_sep_char.setObjectName("format_group_3_sep_char")
        self.gridLayout.addWidget(self.format_group_3_sep_char, 3, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_4_sep_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_4_sep_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_4_sep_char.setObjectName("format_group_4_sep_char")
        self.gridLayout.addWidget(self.format_group_4_sep_char, 4, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_2_end_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_2_end_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_2_end_char.setObjectName("format_group_2_end_char")
        self.gridLayout.addWidget(self.format_group_2_end_char, 2, 3, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_3_end_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_3_end_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_3_end_char.setObjectName("format_group_3_end_char")
        self.gridLayout.addWidget(self.format_group_3_end_char, 3, 3, 1, 1, QtCore.Qt.AlignHCenter)
        self.format_group_4_end_char = QtWidgets.QLineEdit(self.gb_group_settings)
        self.format_group_4_end_char.setMaximumSize(QtCore.QSize(50, 16777215))
        self.format_group_4_end_char.setObjectName("format_group_4_end_char")
        self.gridLayout.addWidget(self.format_group_4_end_char, 4, 3, 1, 1, QtCore.Qt.AlignHCenter)
        self.verticalLayout_6.addLayout(self.gridLayout)
        self.verticalLayout_2.addWidget(self.gb_group_settings)
        self.gb_examples = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.gb_examples.setObjectName("gb_examples")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.gb_examples)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.example_instruments = QtWidgets.QLabel(self.gb_examples)
        self.example_instruments.setText("")
        self.example_instruments.setObjectName("example_instruments")
        self.verticalLayout_4.addWidget(self.example_instruments)
        self.example_vocals = QtWidgets.QLabel(self.gb_examples)
        self.example_vocals.setText("")
        self.example_vocals.setObjectName("example_vocals")
        self.verticalLayout_4.addWidget(self.example_vocals)
        self.verticalLayout_2.addWidget(self.gb_examples)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.retranslateUi(FormatPerformerTagsOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(FormatPerformerTagsOptionsPage)

    def retranslateUi(self, FormatPerformerTagsOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        FormatPerformerTagsOptionsPage.setWindowTitle(_translate("FormatPerformerTagsOptionsPage", "Form"))
        self.gb_description.setTitle(_translate("FormatPerformerTagsOptionsPage", "Format Performer Tags"))
        self.format_description.setText(_translate("FormatPerformerTagsOptionsPage", "<html><head/><body><p>These settings will determine the format for any <span style=\" font-weight:600;\">Performer</span> tags prepared. The format is divided into six parts: the performer; the instrument or &quot;vocals&quot;; and four user selectable sections for the extra information. This is set out as:</p><p align=\"center\"><span style=\" font-weight:600;\">[Section 1]</span>Instrument/Vocals<span style=\" font-weight:600;\">[Section 2][Section 3]</span>: Performer<span style=\" font-weight:600;\">[Section 4]</span></p><p>You can select the section in which each of the extra information words appears.</p><p>For each of the sections you can select the starting character(s), the character(s) separating entries, and the ending character(s). Note that leading or trailing spaces must be included in the settings and will not be automatically added. If no separator characters are entered, the items within a section will be automatically separated by a single space.</p><p>Please visit the repository on GitHub for <a href=\"https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/format_performer_tags/docs/README.md\"><span style=\" text-decoration: underline; color:#0000ff;\">additional information</span></a>.</p></body></html>"))
        self.gb_word_groups.setTitle(_translate("FormatPerformerTagsOptionsPage", "Keyword Sections Assignment"))
        self.group_additonal.setTitle(_translate("FormatPerformerTagsOptionsPage", "Keyword: additional"))
        self.additional_rb_1.setText(_translate("FormatPerformerTagsOptionsPage", "1"))
        self.additional_rb_2.setText(_translate("FormatPerformerTagsOptionsPage", "2"))
        self.additional_rb_3.setText(_translate("FormatPerformerTagsOptionsPage", "3"))
        self.additional_rb_4.setText(_translate("FormatPerformerTagsOptionsPage", "4"))
        self.group_guest.setTitle(_translate("FormatPerformerTagsOptionsPage", "Keyword: guest"))
        self.guest_rb_1.setText(_translate("FormatPerformerTagsOptionsPage", "1"))
        self.guest_rb_2.setText(_translate("FormatPerformerTagsOptionsPage", "2"))
        self.guest_rb_3.setText(_translate("FormatPerformerTagsOptionsPage", "3"))
        self.guest_rb_4.setText(_translate("FormatPerformerTagsOptionsPage", "4"))
        self.group_solo.setTitle(_translate("FormatPerformerTagsOptionsPage", "Keyword: solo"))
        self.solo_rb_1.setText(_translate("FormatPerformerTagsOptionsPage", "1"))
        self.solo_rb_2.setText(_translate("FormatPerformerTagsOptionsPage", "2"))
        self.solo_rb_3.setText(_translate("FormatPerformerTagsOptionsPage", "3"))
        self.solo_rb_4.setText(_translate("FormatPerformerTagsOptionsPage", "4"))
        self.group_vocals.setTitle(_translate("FormatPerformerTagsOptionsPage", "All vocal type keywords"))
        self.vocals_rb_1.setText(_translate("FormatPerformerTagsOptionsPage", "1"))
        self.vocals_rb_2.setText(_translate("FormatPerformerTagsOptionsPage", "2"))
        self.vocals_rb_3.setText(_translate("FormatPerformerTagsOptionsPage", "3"))
        self.vocals_rb_4.setText(_translate("FormatPerformerTagsOptionsPage", "4"))
        self.gb_group_settings.setTitle(_translate("FormatPerformerTagsOptionsPage", "Section Display Settings"))
        self.label_2.setText(_translate("FormatPerformerTagsOptionsPage", "Section 2"))
        self.label_3.setText(_translate("FormatPerformerTagsOptionsPage", "Section 3"))
        self.format_group_1_start_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.label.setText(_translate("FormatPerformerTagsOptionsPage", "Section 1"))
        self.label_4.setText(_translate("FormatPerformerTagsOptionsPage", "Section 4"))
        self.format_group_1_sep_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_1_end_char.setText(_translate("FormatPerformerTagsOptionsPage", " "))
        self.format_group_1_end_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.label_5.setText(_translate("FormatPerformerTagsOptionsPage", "Start Char(s)"))
        self.label_6.setText(_translate("FormatPerformerTagsOptionsPage", "Sep Char(s)"))
        self.label_7.setText(_translate("FormatPerformerTagsOptionsPage", "End Char(s)"))
        self.format_group_2_start_char.setText(_translate("FormatPerformerTagsOptionsPage", ", "))
        self.format_group_2_start_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_3_start_char.setText(_translate("FormatPerformerTagsOptionsPage", " ("))
        self.format_group_3_start_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_4_start_char.setText(_translate("FormatPerformerTagsOptionsPage", " ("))
        self.format_group_4_start_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_2_sep_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_3_sep_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_4_sep_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_2_end_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_3_end_char.setText(_translate("FormatPerformerTagsOptionsPage", ")"))
        self.format_group_3_end_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.format_group_4_end_char.setText(_translate("FormatPerformerTagsOptionsPage", ")"))
        self.format_group_4_end_char.setPlaceholderText(_translate("FormatPerformerTagsOptionsPage", "(blank)"))
        self.gb_examples.setTitle(_translate("FormatPerformerTagsOptionsPage", "Examples"))



PLUGIN_NAME = 'Release Type'
PLUGIN_AUTHOR = 'Elliot Chance'
PLUGIN_DESCRIPTION = 'Appends information to EPs and Singles'
PLUGIN_VERSION = '1.4'
PLUGIN_API_VERSIONS = ["0.9.0", "0.10", "0.15", "2.0"]

from picard.metadata import register_album_metadata_processor

RELEASE_TYPE_MAPPING = {
    "ep": " EP",
    "single": " (single)"
}


def add_release_type(tagger, metadata, release):

    # make sure "EP" (or "single", ...) is not already a word in the name
    words = metadata["album"].lower().split(" ")
    for word in ["ep", "e.p.", "single", "(single)"]:
        if word in words:
            return

    # check release type
    for releasetype, text in RELEASE_TYPE_MAPPING.items():
        if (metadata["~primaryreleasetype"] == releasetype) or (
                metadata["releasetype"].startswith(releasetype)):
            rs = text
            break
    else:
        rs = ""

  # append title
    metadata["album"] = metadata["album"] + rs

register_album_metadata_processor(add_release_type)



# -*- coding: utf-8 -*-

from __future__ import print_function
import ast

KNOWN_DATA = [
    'PLUGIN_NAME',
    'PLUGIN_AUTHOR',
    'PLUGIN_VERSION',
    'PLUGIN_API_VERSIONS',
    'PLUGIN_LICENSE',
    'PLUGIN_LICENSE_URL',
    'PLUGIN_DESCRIPTION',
]


def get_plugin_data(filepath):
    """Parse a python file and return a dict with plugin metadata"""
    data = {}
    with open(filepath, 'r', encoding='utf-8') as plugin_file:
        source = plugin_file.read()
        try:
            root = ast.parse(source, filepath)
        except:
            print("Cannot parse " + filepath)
            raise
        for node in ast.iter_child_nodes(root):
            if isinstance(node, ast.Assign) and len(node.targets) == 1:
                target = node.targets[0]
                if (isinstance(target, ast.Name)
                    and isinstance(target.ctx, ast.Store)
                        and target.id in KNOWN_DATA):
                    name = target.id.replace('PLUGIN_', '', 1).lower()
                    if name not in data:
                        try:
                            data[name] = ast.literal_eval(node.value)
                        except ValueError:
                            print('Cannot evaluate value in '
                                  + filepath + ':' +
                                  ast.dump(node))
        return data



PLUGIN_NAME = "Collect Album Artists"
PLUGIN_AUTHOR = "johbi"
PLUGIN_DESCRIPTION = "Adds a context menu shortcut to collect all track artists from a release and format them as the releases album artist."
PLUGIN_VERSION = '0.1'
PLUGIN_API_VERSIONS = ['2.1', '2.2']
PLUGIN_LICENSE = "GPL-3.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl.txt"

from picard.album import Album
from picard.ui.itemviews import BaseAction, register_album_action

class CollectArtists(BaseAction):
    NAME = 'Replace release artist with &track artists...'

    def callback(self, objs):
        for album in objs:
            if isinstance(album, Album):
                trackartists = []
                for track in album.tracks:
                    if "artists" in track.metadata:
                        artists = track.metadata.getall("artists")
                    elif "artist" in track.metadata:
                        artists = track.metadata.getall("artist")
                    else:
                        continue

                    for artist in artists:
                        if artist not in trackartists:
                            trackartists.append(artist)

                if len(trackartists) >= 2:
                    albumartist = (", ").join(trackartists[:-1]) + ' & ' + trackartists[-1]
                elif len(trackartists) == 1:
                    albumartist = trackartists[0]
                else:
                    self.tagger.window.set_statusbar_message("Could not find any artists for the album: \"" + album.metadata['album'] + "\"")
                    continue
                album.metadata.set("albumartist", albumartist)

                for track in album.tracks:
                    track.metadata.set("albumartist", albumartist)

                    for files in track.linked_files:
                        track.update_file_metadata(files)

                album.update()

register_album_action(CollectArtists())



from picard.collection import Collection, user_collections
from picard.ui.options import OptionsPage, register_options_page

from picard.plugins.add_to_collection import settings
from picard.plugins.add_to_collection.manifest import PLUGIN_NAME
from picard.plugins.add_to_collection.ui_add_to_collection_options import (
    Ui_AddToCollectionOptions,
)
from picard.plugins.add_to_collection.override_module import override_module


class AddToCollectionOptionsPage(OptionsPage):
    NAME = "add-to-collection"
    TITLE = PLUGIN_NAME
    PARENT = "plugins"

    options = [settings.collection_id_option()]

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.ui = Ui_AddToCollectionOptions()
        self.ui.setupUi(self)

    def load(self) -> None:
        self.set_collection_name(settings.collection_id())

    def save(self) -> None:
        settings.set_collection_id(self.ui.collection_name.currentData())

    def set_collection_name(self, value: str) -> None:
        self.ui.collection_name.clear()
        collection: Collection
        for collection in user_collections.values():
            self.ui.collection_name.addItem(collection.name, collection.id)
        idx = self.ui.collection_name.findData(value)
        if idx != -1:
            self.ui.collection_name.setCurrentIndex(idx)


def register_options() -> None:
    with override_module(AddToCollectionOptionsPage):
        register_options_page(AddToCollectionOptionsPage)



import os
from itertools import chain
from test.plugin_test_case import PluginTestCase
from typing import Union
from unittest.mock import ANY, patch

from picard.collection import Collection, user_collections
from picard.file import File
from picard.file import _file_post_save_processors as post_save_processors
from picard.ui.options import _pages as option_pages


class TestAddToCollection(PluginTestCase):
    SAVE_FILE_SETTINGS = {
        "dont_write_tags": True,
        "rename_files": False,
        "move_files": False,
        "delete_empty_dirs": False,
        "save_images_to_files": False,
        "clear_existing_tags": False,
        "compare_ignore_tags": [],
    }

    def install_plugin(self) -> None:
        self.plugin = self._test_plugin_install(
            "Add to Collection", "add_to_collection"
        )

    def create_file(self, file_name: str, album_id: Union[str, None] = None) -> File:
        file_path = os.path.join(self.tmp_directory, file_name)
        file = File(file_path)
        if album_id:
            file.metadata["musicbrainz_albumid"] = album_id
        self.tagger.files[file_path] = file
        return file

    def tearDown(self) -> None:
        user_collections.clear()
        return super().tearDown()

    def test_hooks_installed(self) -> None:
        self.install_plugin()

        self.assertIn(
            self.plugin.post_save_processor.post_save_processor,
            list(chain.from_iterable(post_save_processors.functions.values())),
        )

        self.assertIn(
            self.plugin.options.AddToCollectionOptionsPage, list(option_pages)
        )

    def test_file_save(self) -> None:
        self.install_plugin()

        self.set_config_values(
            setting={
                **self.SAVE_FILE_SETTINGS,
                "add_to_collection_id": "test_collection_id",
            }
        )

        file = self.create_file(file_name="test.mp3", album_id="test_album_id")
        user_collections["test_collection_id"] = Collection(
            collection_id="test_collection_id", name="Test Collection", size=0
        )
        with patch("picard.collection.Collection.add_releases") as add_releases:
            file.save()
            add_releases.assert_called_with(set(["test_album_id"]), callback=ANY)

    def test_two_files_save(self) -> None:
        self.install_plugin()

        self.set_config_values(
            setting={
                **self.SAVE_FILE_SETTINGS,
                "add_to_collection_id": "test_collection_id",
            }
        )

        collection = Collection(
            collection_id="test_collection_id", name="Test Collection", size=0
        )
        user_collections["test_collection_id"] = collection
        with patch(
            "picard.collection.Collection.add_releases",
            side_effect=lambda ids, callback: collection.releases.update(ids),
        ) as add_releases:
            for i in range(2):
                file = self.create_file(
                    file_name=f"test{i}.mp3", album_id="test_album_id"
                )
                file.save()
            # only added once
            add_releases.assert_called_once_with(set(["test_album_id"]), callback=ANY)

    def test_no_collection_id_setting(self) -> None:
        self.install_plugin()

        self.set_config_values(setting=self.SAVE_FILE_SETTINGS)

        file = self.create_file(file_name="test.mp3", album_id="test_album_id")
        user_collections["test_collection_id"] = Collection(
            collection_id="test_collection_id", name="Test Collection", size=0
        )
        with patch("picard.collection.Collection.add_releases") as add_releases:
            file.save()
            add_releases.assert_not_called()

    def test_no_user_collection(self) -> None:
        self.install_plugin()

        self.set_config_values(
            setting={
                **self.SAVE_FILE_SETTINGS,
                "add_to_collection_id": "test_collection_id",
            }
        )

        file = self.create_file(file_name="test.mp3", album_id="test_album_id")
        with patch("picard.collection.Collection.add_releases") as add_releases:
            file.save()
            add_releases.assert_not_called()

    def test_no_release(self) -> None:
        self.install_plugin()

        self.set_config_values(
            setting={
                **self.SAVE_FILE_SETTINGS,
                "add_to_collection_id": "test_collection_id",
            }
        )

        file = self.create_file(file_name="test.mp3")
        with patch("picard.collection.Collection.add_releases") as add_releases:
            file.save()
            add_releases.assert_not_called()



# -*- coding: utf-8 -*-
#This is not my work. I just changed the language to Greek.
#All the credits goes to the original coder.

# This is the Decode Greek plugin for MusicBrainz Picard.
# Copyright (C) 2015 aeontech
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
PLUGIN_NAME = "Decode Cyrillic Greek"
PLUGIN_AUTHOR = "aeontech, Lefteris NeNpO"
PLUGIN_VERSION = "1.3"
PLUGIN_API_VERSIONS = ["1.0", "2.0"]
PLUGIN_LICENSE = "MIT"
PLUGIN_LICENSE_URL = "https://opensource.org/licenses/MIT"
PLUGIN_DESCRIPTION = '''
This plugin helps you quickly convert mis-encoded Greek Windows-1253 tags
to proper UTF-8 encoded strings. If your track/album names look something like
"Àëèñà â ñò›àíå ÷óäåñ", run this plugin from the context menu
before running the "Lookup" or "Scan" tools
'''
from picard import log
from picard.cluster import Cluster
from picard.ui.itemviews import BaseAction, register_cluster_action

_decode_tags = [
    'title',
    'albumartist',
    'albumartistsort',
    'artist',
    'artistsort',
    'album',
    'comment:',
    'comment:ID3v1 Comment'
]
class DecodeGreek(BaseAction):
    NAME = "Unmangle Greek metadata"
    def unmangle(self, tag, value):
        try:
            log.debug("%s: %s => %r" % (PLUGIN_NAME, value, value.encode('latin1')))
            unmangled_value = value.encode('latin1').decode('cp1253')
        except UnicodeError:
            unmangled_value = value
            log.debug("%s: could not unmangle tag %s; original value: %s" % (PLUGIN_NAME, tag, value))
        return unmangled_value
    def callback(self, objs):
        for cluster in objs:
            if not isinstance(cluster, Cluster):
                continue
            for tag in _decode_tags:
                if not (tag in cluster.metadata):
                    continue
                cluster.metadata[tag] = self.unmangle(tag, cluster.metadata[tag])
            log.debug("cluster name is %s by %s" % (cluster.metadata['album'], cluster.metadata['albumartist']))
            for file in cluster.files:
                log.debug("%s: Trying to unmangle file - original metadata %s" % (PLUGIN_NAME, file.orig_metadata))
                for tag in _decode_tags:
                    if not (tag in file.metadata):
                        continue
                    unmangled_tag = self.unmangle(tag, file.metadata[tag])
                    file.orig_metadata[tag] = unmangled_tag
                    file.metadata[tag] = unmangled_tag
                    file.orig_metadata.changed = True
                    file.metadata.changed = True
                    file.update(signal=True)
            cluster.update()
register_cluster_action(DecodeGreek())



#!/usr/bin/python
# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

PLUGIN_NAME = "Generate M3U playlist"
PLUGIN_AUTHOR = "Francis Chin, Sambhav Kothari, Chris Hylen"
PLUGIN_DESCRIPTION = """Generate an Extended M3U playlist (.m3u8 file, UTF8
encoded text). Relative pathnames are used where audio files are in the same
directory as the playlist, otherwise absolute (full) pathnames are used."""
PLUGIN_VERSION = "1.2.1"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

import os.path

from PyQt5 import QtCore, QtWidgets
from picard import log
from picard.const import VARIOUS_ARTISTS_ID
from picard.util import find_existing_path, encode_filename
from picard.ui.itemviews import BaseAction, register_album_action


_debug_level = 0

def get_safe_filename(filename):
    _valid_chars = " .,_-:+&!()"
    _safe_filename = "".join(
        c if (c.isalnum() or c in _valid_chars) else "_" for c in filename
    ).rstrip()
    return _safe_filename


class PlaylistEntry(list):

    def __init__(self, playlist, index):
        list.__init__(self)
        self.playlist = playlist
        self.index = index

    def add(self, entry_row):
        self.append(entry_row + "\n")


class Playlist(object):

    def __init__(self, filename):
        self.filename = filename
        self.entries = []
        self.headers = []

    def add_header(self, header):
        self.headers.append(header + "\n")

    def write(self):
        b_lines = []
        for header in self.headers:
            b_lines.append(header.encode("utf-8"))
        for entry in self.entries:
            for row in entry:
                b_lines.append(row.encode("utf-8"))
        with open(encode_filename(self.filename), "wb") as f:
            f.writelines(b_lines)


class GeneratePlaylist(BaseAction):
    NAME = "Generate &Playlist..."

    def callback(self, objs):
        # Find common path of all files to default where to save playlist
        files = []
        for album in objs:
            for track in album.tracks:
                if track.linked_files:
                    files.append(track.linked_files[0].filename)

        try:
            current_directory = os.path.commonpath(files)
        except ValueError:
            current_directory = ""

        # Default playlist filename set as "%albumartist% - %album%.m3u8",
        # except where "Various Artists" is suppressed
        if _debug_level > 1:
            log.debug("{}: VARIOUS_ARTISTS_ID is {}, musicbrainz_albumartistid is {}".format(
                    PLUGIN_NAME, VARIOUS_ARTISTS_ID,
                    objs[0].metadata["musicbrainz_albumartistid"]))
        if objs[0].metadata["musicbrainz_albumartistid"] != VARIOUS_ARTISTS_ID:
            default_filename = get_safe_filename(
                objs[0].metadata["albumartist"] + " - "
                + objs[0].metadata["album"] + ".m3u8"
            )
        else:
            default_filename = get_safe_filename(
                objs[0].metadata["album"] + ".m3u8"
            )
        if _debug_level > 1:
            log.debug("{}: default playlist filename sanitized to {}".format(
                    PLUGIN_NAME, default_filename))
        filename, selected_format = QtWidgets.QFileDialog.getSaveFileName(
            None, "Save new playlist",
            os.path.join(current_directory, default_filename),
            "Playlist (*.m3u8 *.m3u)"
        )
        if filename:
            playlist = Playlist(filename)
            playlist.add_header("#EXTM3U")

            for album in objs:
                for track in album.tracks:
                    if track.linked_files:
                        entry = PlaylistEntry(playlist, len(playlist.entries))
                        playlist.entries.append(entry)

                        # M3U EXTINF row
                        track_length_seconds = int(round(track.metadata.length / 1000.0))
                        # EXTINF format assumed to be fixed as follows:
                        entry.add("#EXTINF:{duration:d},{artist} - {title}".format(
                            duration=track_length_seconds,
                            artist=track.metadata["artist"],
                            title=track.metadata["title"]
                            )
                        )

                        # M3U URL row - assumes only one file per track
                        audio_filename = track.linked_files[0].filename
                        if _debug_level > 1:
                            for i, file in enumerate(track.linked_files):
                                log.debug("{}: linked_file {}: {}".format(
                                    PLUGIN_NAME, i, str(file)))
                        # If playlist is in same directory as audio files, then use
                        # local (relative) pathname, otherwise use absolute pathname
                        if _debug_level > 1:
                            log.debug("{}: audio_filename: {}, selected dir: {}".format(
                                    PLUGIN_NAME, audio_filename, os.path.dirname(filename)))

                        try:
                            audio_filename = os.path.relpath(audio_filename, os.path.dirname(filename))
                        except ValueError:
                            pass
                        entry.add(str(audio_filename))

            playlist.write()


register_album_action(GeneratePlaylist())



# -*- coding: utf-8 -*-
# Copyright 2007 Javier Kohen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation

PLUGIN_NAME = "Title Case"
PLUGIN_AUTHOR = "Javier Kohen, Sambhav Kothari"
PLUGIN_DESCRIPTION = "Capitalize First Character In Every Word Of A Title"
PLUGIN_VERSION = "1.0.2"
PLUGIN_API_VERSIONS = ['2.0']
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

import unicodedata
from picard.plugin import PluginPriority


def iswbound(char):
    """Returns whether the given character is a word boundary."""
    category = unicodedata.category(char)
    # If it's a space separator or punctuation
    return 'Zs' == category or 'Sk' == category or 'P' == category[0]


def utitle(string):
    """Title-case a string using a less destructive method than str.title."""
    new_string = string[0].capitalize()
    cap = False
    for i in range(1, len(string)):
        s = string[i]
        # Special case apostrophe in the middle of a word.
        if s in "’'" and string[i - 1].isalpha():
            cap = False
        elif iswbound(s):
            cap = True
        elif cap and s.isalpha():
            cap = False
            s = s.capitalize()
        else:
            cap = False
        new_string += s
    return new_string


def title(string):
    """Title-case a string using a less destructive method than str.title."""
    if not string:
        return ""
    # if the string is all uppercase, lowercase it - Erich/Javier
    #   Lots of Japanese songs use entirely upper-case English titles,
    #   so I don't like this change... - JoeW
    #if string == string.upper(): string = string.lower()
    return utitle(str(string))

from picard.metadata import (
    register_track_metadata_processor,
    register_album_metadata_processor,
)


def title_case(tagger, metadata, *args):
    for name, value in metadata.rawitems():
        if name in ["title", "album", "artist"]:
            metadata[name] = [title(x) for x in value]

register_track_metadata_processor(title_case, priority=PluginPriority.LOW)
register_album_metadata_processor(title_case, priority=PluginPriority.LOW)



#!/usr/bin/env python
# coding: utf-8
import unittest

from picard.metadata import Metadata
from picard.script import ScriptParser
from plugins.keep.keep import keep  # noqa: F401 pylint: disable=unused-import


class TestKeep(unittest.TestCase):
    def setUp(self):
        self.parser = ScriptParser()

    def test_keep_simple(self):
        meta = Metadata(
            {
                "foo": "foo",
                "bar": "bar"
            })

        sc = """$keep(bar)"""
        self.parser.eval(sc, meta)
        self.assertEqual(meta["bar"], "bar")
        self.assertEqual(len(meta.keys()), 1)
        self.assertNotIn("foo", meta)

    def test_keep_mbid(self):
        meta = Metadata(
            {
                "foo": "foo",
                "bar": "bar",
                "musicbrainz_albumid": "albumid",
            })

        sc = """$keep(bar)"""
        self.parser.eval(sc, meta)
        self.assertEqual(meta["musicbrainz_albumid"], "albumid")
        self.assertEqual(len(meta.keys()), 2)

    def test_keep_nonfiletags(self):
        meta = Metadata(
            {
                "foo": "foo",
                "bar": "bar",
                "~baz": "baz",
            })

        sc = """$keep(bar)"""
        self.parser.eval(sc, meta)
        self.assertEqual(meta["~baz"], "baz")
        self.assertEqual(len(meta.keys()), 2)

    def _description_test(self, tagname):
        meta = Metadata(
            {
                "foo": "foo",
                "bar": "bar",
                tagname: tagname
            })

        tag_without_description = tagname.split(":")[0]
        sc = """$keep({tag})""".format(tag=tag_without_description)
        self.parser.eval(sc, meta)
        self.assertEqual(meta[tagname], tagname)
        self.assertEqual(len(meta.keys()), 1)

    def test_keep_performer(self):
        self._description_test("performer:vocal")
        self._description_test("performer:")

    def test_keep_lyrics(self):
        self._description_test("lyrics:foo")
        self._description_test("lyrics:")

    def test_keep_comment(self):
        self._description_test("comment:foo")
        self._description_test("comment:")

    def test_keep_with_description(self):
        meta = Metadata(
            {
                "foo": "foo",
                "bar": "bar",
                "performer:vocal": "performer:vocal",
                "performer:guitar": "performer:guitar"
            })

        sc = """$keep(performer:vocal)"""
        self.parser.eval(sc, meta)
        self.assertEqual(meta["performer:vocal"], "performer:vocal")
        self.assertEqual(len(meta.keys()), 1)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/lastfm/ui_options_lastfm.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LastfmOptionsPage(object):
    def setupUi(self, LastfmOptionsPage):
        LastfmOptionsPage.setObjectName("LastfmOptionsPage")
        LastfmOptionsPage.resize(305, 317)
        self.vboxlayout = QtWidgets.QVBoxLayout(LastfmOptionsPage)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.rename_files = QtWidgets.QGroupBox(LastfmOptionsPage)
        self.rename_files.setObjectName("rename_files")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.rename_files)
        self.vboxlayout1.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout1.setSpacing(2)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.use_track_tags = QtWidgets.QCheckBox(self.rename_files)
        self.use_track_tags.setObjectName("use_track_tags")
        self.vboxlayout1.addWidget(self.use_track_tags)
        self.use_artist_tags = QtWidgets.QCheckBox(self.rename_files)
        self.use_artist_tags.setObjectName("use_artist_tags")
        self.vboxlayout1.addWidget(self.use_artist_tags)
        self.vboxlayout.addWidget(self.rename_files)
        self.rename_files_2 = QtWidgets.QGroupBox(LastfmOptionsPage)
        self.rename_files_2.setObjectName("rename_files_2")
        self.vboxlayout2 = QtWidgets.QVBoxLayout(self.rename_files_2)
        self.vboxlayout2.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout2.setSpacing(2)
        self.vboxlayout2.setObjectName("vboxlayout2")
        self.ignore_tags_2 = QtWidgets.QLabel(self.rename_files_2)
        self.ignore_tags_2.setObjectName("ignore_tags_2")
        self.vboxlayout2.addWidget(self.ignore_tags_2)
        self.ignore_tags = QtWidgets.QLineEdit(self.rename_files_2)
        self.ignore_tags.setObjectName("ignore_tags")
        self.vboxlayout2.addWidget(self.ignore_tags)
        self.hboxlayout = QtWidgets.QHBoxLayout()
        self.hboxlayout.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout.setSpacing(6)
        self.hboxlayout.setObjectName("hboxlayout")
        self.ignore_tags_4 = QtWidgets.QLabel(self.rename_files_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ignore_tags_4.sizePolicy().hasHeightForWidth())
        self.ignore_tags_4.setSizePolicy(sizePolicy)
        self.ignore_tags_4.setObjectName("ignore_tags_4")
        self.hboxlayout.addWidget(self.ignore_tags_4)
        self.join_tags = QtWidgets.QComboBox(self.rename_files_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.join_tags.sizePolicy().hasHeightForWidth())
        self.join_tags.setSizePolicy(sizePolicy)
        self.join_tags.setEditable(True)
        self.join_tags.setObjectName("join_tags")
        self.join_tags.addItem("")
        self.join_tags.setItemText(0, "")
        self.join_tags.addItem("")
        self.join_tags.addItem("")
        self.hboxlayout.addWidget(self.join_tags)
        self.vboxlayout2.addLayout(self.hboxlayout)
        self.hboxlayout1 = QtWidgets.QHBoxLayout()
        self.hboxlayout1.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout1.setSpacing(6)
        self.hboxlayout1.setObjectName("hboxlayout1")
        self.label_4 = QtWidgets.QLabel(self.rename_files_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setObjectName("label_4")
        self.hboxlayout1.addWidget(self.label_4)
        self.min_tag_usage = QtWidgets.QSpinBox(self.rename_files_2)
        self.min_tag_usage.setMaximum(100)
        self.min_tag_usage.setObjectName("min_tag_usage")
        self.hboxlayout1.addWidget(self.min_tag_usage)
        self.vboxlayout2.addLayout(self.hboxlayout1)
        self.vboxlayout.addWidget(self.rename_files_2)
        spacerItem = QtWidgets.QSpacerItem(263, 21, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem)
        self.label_4.setBuddy(self.min_tag_usage)

        self.retranslateUi(LastfmOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(LastfmOptionsPage)
        LastfmOptionsPage.setTabOrder(self.use_track_tags, self.ignore_tags)

    def retranslateUi(self, LastfmOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.rename_files.setTitle(_translate("LastfmOptionsPage", "Last.fm"))
        self.use_track_tags.setText(_translate("LastfmOptionsPage", "Use track tags"))
        self.use_artist_tags.setText(_translate("LastfmOptionsPage", "Use artist tags"))
        self.rename_files_2.setTitle(_translate("LastfmOptionsPage", "Tags"))
        self.ignore_tags_2.setText(_translate("LastfmOptionsPage", "Ignore tags:"))
        self.ignore_tags_4.setText(_translate("LastfmOptionsPage", "Join multiple tags with:"))
        self.join_tags.setItemText(1, _translate("LastfmOptionsPage", " / "))
        self.join_tags.setItemText(2, _translate("LastfmOptionsPage", ", "))
        self.label_4.setText(_translate("LastfmOptionsPage", "Minimal tag usage:"))
        self.min_tag_usage.setSuffix(_translate("LastfmOptionsPage", " %"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015, 2019 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Paper CD case'
PLUGIN_AUTHOR = 'Philipp Wolfer, Sambhav Kothari'
PLUGIN_DESCRIPTION = ('Create a paper CD case from an album or cluster '
                      'using http://papercdcase.com')
PLUGIN_VERSION = "1.2.1"
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


from PyQt5.QtCore import QUrl, QUrlQuery
from picard.album import Album
from picard.cluster import Cluster
from picard.ui.itemviews import (
    BaseAction,
    register_album_action,
    register_cluster_action,
)
from picard.util import webbrowser2
from picard.util import textencoding


PAPERCDCASE_URL = 'http://papercdcase.com/advanced.php'
URLENCODE_ASCII_CHARS = [ord(' '), ord('%'), ord('&'), ord('/'), ord('?')]


def urlencode(s):
    chars = []
    for c in s:
        n = ord(c)
        if (n < 128 or n > 255) and n not in URLENCODE_ASCII_CHARS:
            chars.append(c)
        else:
            chars.append('%' + hex(n)[2:].upper())
    return ''.join(chars)


def build_papercdcase_url(artist, album, tracks):
    url = QUrl(PAPERCDCASE_URL)
    # papercdcase.com does not deal well with unicode characters :(
    url_query = QUrlQuery()
    url_query.addQueryItem('artist',
                           urlencode(textencoding.asciipunct(artist)))
    url_query.addQueryItem('title',
                           urlencode(textencoding.asciipunct(album)))
    i = 1
    for track in tracks:
        url_query.addQueryItem('track' + str(i),
                               urlencode(textencoding.asciipunct(track)))
        i += 1
    url.setQuery(url_query)
    return url.toString(QUrl.FullyEncoded)


class PaperCdCase(BaseAction):
    NAME = 'Create paper CD case'

    def callback(self, objs):
        for obj in objs:
            if isinstance(obj, Album) or isinstance(obj, Cluster):
                artist = obj.metadata['albumartist']
                album = obj.metadata['album']
                if isinstance(obj, Album):
                    tracks = [track.metadata['title'] for track in obj.tracks]
                else:
                    tracks = [file.metadata['title'] for file in obj.files]
                url = build_papercdcase_url(artist, album, tracks)
                webbrowser2.open(url)


paperCdCase = PaperCdCase()
register_album_action(paperCdCase)
register_cluster_action(paperCdCase)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'MOD files'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = (
    'Support for loading and renaming various tracker module formats '
    '(.mod, .xm, .it, .mptm, .ahx, .mtm, .med, .s3m, .ult, .699, .okt). '
    'There is limited support for writing the title tag as track name for '
    'some formats.'
)
PLUGIN_VERSION = "0.1"
PLUGIN_API_VERSIONS = ["2.8"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from dataclasses import dataclass
from enum import Enum
from io import RawIOBase
import struct

from mutagen._util import resize_bytes

from picard import log
from picard.file import File
from picard.formats import register_format
from picard.metadata import Metadata
from picard.util.textencoding import asciipunct


class FieldAccess(Enum):
    READ_ONLY = 0
    READ_WRITE = 1


@dataclass
class StaticField:
    name: str
    offset: int
    length: int
    access: FieldAccess
    fillchar: str = ' '


class MagicBytes(bytes):
    offset: int = 0

    def __new__(cls, value, offset: int = 0):
        self = super().__new__(cls, value)
        self.offset = offset
        return self


class ModuleFile(File):
    # Allows to specify a magic number to identify the file format.
    # Re-implement _ensure_format for more complex cases.
    _magic_bytes: tuple[MagicBytes] = None

    # Specify the encoding for the format.
    # There is not much info about encoding, most files seem to be limited
    # to ASCII. But cp850 seems to be a solid default for the few files that
    # use 8-bit characters.
    _encoding = 'cp850'  # or cp437?

    # List of StaticField. Most fields in mod files
    # have static position and fixed size.
    _static_text_fields: tuple[StaticField] = ()

    @classmethod
    def supports_tag(cls, name: str) -> bool:
        return name in {field.name for field in cls._static_text_fields}

    def _load(self, filename: str) -> Metadata:
        log.debug('Loading file %r', filename)
        metadata = Metadata()
        self._add_path_to_metadata(metadata)
        metadata['~format'] = self.NAME
        with open(filename, 'rb') as f:
            magic = self._ensure_format(f)
            self._parse_file(f, metadata, magic)
        return metadata

    def _save(self, filename: str, metadata: Metadata):
        log.debug('Saving file %r', filename)
        with open(filename, 'rb+') as f:
            self._ensure_format(f)
            self._write_file(f, metadata)

    def _ensure_format(self, f: RawIOBase) -> MagicBytes:
        if not self._magic_bytes:
            raise NotImplementedError('_magic_bytes not set or method not implemented')
        for magic in self._magic_bytes:
            if self._magic_matches(f, magic):
                return magic
        # None of the magic byte sequences matched, fail loading
        raise ValueError('Not a %s file' % self.NAME)

    def _magic_matches(self, f: RawIOBase, magic: MagicBytes) -> bool:
        f.seek(magic.offset)
        file_id = f.read(len(magic))
        return file_id == magic

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        for field in self._static_text_fields:
            f.seek(field.offset)
            metadata[field.name] = self._decode_text(f.read(field.length))

    def _write_file(self, f: RawIOBase, metadata: Metadata):
        for field in self._static_text_fields:
            if field.access == FieldAccess.READ_WRITE and not field.name.startswith('~'):
                f.seek(field.offset)
                f.write(self._encode_text(metadata[field.name], field.length, field.fillchar))

    def _decode_text(self, data: bytes) -> str:
        return data.decode(self._encoding, errors='replace').strip().strip('\0')

    def _encode_text(self, text: str, length: int = None, fillchar: str = ' ') -> bytes:
        if length:
            text = text[:length].ljust(length, fillchar)
        return asciipunct(text).encode(self._encoding, errors='replace')


class MODFile(ModuleFile):
    EXTENSIONS = ['.mod']
    NAME = 'MOD'

    # https://www.ocf.berkeley.edu/~eek/index.html/tiny_examples/ptmod/ap12.html
    _magic_bytes = (MagicBytes(b'M\x2eK\x2e', offset=1080),)
    _static_text_fields = (
        StaticField('title', 0, 20, FieldAccess.READ_WRITE),
    )


class ExtendedModuleFile(ModuleFile):
    EXTENSIONS = ['.xm']
    NAME = 'Extended Module'

    # https://github.com/milkytracker/MilkyTracker/blob/master/resources/reference/xm-form.txt
    _magic_bytes = (MagicBytes(b'Extended Module: '),)
    _static_text_fields = (
        StaticField('title', 17, 20, FieldAccess.READ_WRITE),
        StaticField('encodedby', 38, 20, FieldAccess.READ_WRITE),
    )

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        super()._parse_file(f, metadata, magic)
        # OpenMPT seems to use iso-8859-1 encoding.
        if metadata['encodedby'].startswith('OpenMPT'):
            self._encoding = 'iso-8859-1'
            super()._parse_file(f, metadata, magic)
        f.seek(68)
        metadata['~channels'] = struct.unpack('<h', f.read(2))[0]


class ImpulseTrackerFile(ModuleFile):
    EXTENSIONS = ['.it', '.mptm']
    NAME = 'Impulse Tracker'

    # https://fileformats.fandom.com/wiki/Impulse_tracker
    # https://wiki.openmpt.org/Manual:_Module_formats#The_OpenMPT_format_.28.mptm.29
    _magic_bytes = (MagicBytes(b'IMPM'),)
    _static_text_fields = (
        StaticField('title', 4, 26, FieldAccess.READ_WRITE, '\0'),
    )

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        if self._magic_matches(f, MagicBytes(b'OMPT', offset=60)):
            self._encoding = 'iso-8859-1'
            metadata['~format'] = 'OpenMPT'
            # TODO: For OpenMPT enhanced format parse also the author and comment
        super()._parse_file(f, metadata, magic)


class AHXFile(ModuleFile):
    EXTENSIONS = ['.ahx']
    NAME = 'AHX'

    # http://lclevy.free.fr/exotica/ahx/ahxformat.txt
    _magic_bytes = (MagicBytes(b'THX'),)

    @classmethod
    def supports_tag(cls, name: str) -> bool:
        return name in {'title'}

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        self._seek_names_offset(f)
        metadata['title'] = self._decode_text(self._read_string(f))

    def _write_file(self, f: RawIOBase, metadata: Metadata):
        # Write the title (first null terminated string after the samples)
        names_offset = self._seek_names_offset(f)
        old_title = self._read_string(f)
        new_title = self._encode_text(metadata['title'])
        resize_bytes(f, len(old_title), len(new_title), names_offset)
        f.seek(names_offset)
        f.write(new_title)

    def _seek_names_offset(self, f: RawIOBase) -> int:
        f.seek(6)
        len_ = struct.unpack('>H', f.read(2))[0] & 0xfff
        f.seek(10)
        trl, trk, smp, ss = struct.unpack('BBBB', f.read(4))
        samples_offset = 14 + ss*2 + len_*8 + (trk+1)*trl*3
        f.seek(samples_offset)
        self._skip_samples(f, count=smp)
        return f.tell()

    def _skip_samples(self, f: RawIOBase, count: int):
        while count:
            f.seek(f.tell() + 21)
            plen = struct.unpack('B', f.read(1))[0]
            f.seek(f.tell() + plen*4)
            count -= 1

    def _read_string(self, f: RawIOBase) -> bytes:
        """Reads a null terminated string from the current position."""
        result = b''
        char = f.read(1)
        while char and char != b'\0':
            result += char
            char = f.read(1)
        return result


class MEDFile(ModuleFile):
    EXTENSIONS = ['.med']
    NAME = 'MED'

    # https://github.com/dv1/ion_player/blob/master/extern/uade-2.13/amigasrc/players/med/MMD_FileFormat.doc
    _magic_bytes = (
        MagicBytes(b'MMD0'),
        MagicBytes(b'MMD1'),
        MagicBytes(b'MMD2'),
        MagicBytes(b'MMD3'),
    )

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        # TODO: Extract songname
        super()._parse_file(f, metadata, magic)
        metadata['~format'] = '%s (%s)' % (self.NAME, self._decode_text(magic))


class MTMFile(ModuleFile):
    EXTENSIONS = ['.mtm']
    NAME = 'MTM'

    # https://www.fileformat.info/format/mtm/corion.htm
    _magic_bytes = (MagicBytes(b'MTM'),)
    _static_text_fields = (
        StaticField('title', 4, 20, FieldAccess.READ_WRITE, '\0'),
    )


class S3MFile(ModuleFile):
    EXTENSIONS = ['.s3m']
    NAME = 'S3M'

    # https://www.fileformat.info/format/screamtracker/corion.htm
    _magic_bytes = (MagicBytes(b'\x1a', offset=28),)
    _static_text_fields = (
        StaticField('title', 0, 20, FieldAccess.READ_WRITE, '\0'),
        StaticField('encodedby', 20, 8, FieldAccess.READ_WRITE, '\0'),
    )


class ULTFile(ModuleFile):
    EXTENSIONS = ['.ult']
    NAME = 'ULT'

    # http://www.textfiles.com/programming/FORMATS/ultform.pro
    # http://www.textfiles.com/programming/FORMATS/ultform14.pro
    _magic_bytes = (
        MagicBytes(b'MAS_UTrack_V001'),
        MagicBytes(b'MAS_UTrack_V002'),
        MagicBytes(b'MAS_UTrack_V003'),
        MagicBytes(b'MAS_UTrack_V004'),
    )
    _static_text_fields = (
        StaticField('title', 15, 32, FieldAccess.READ_WRITE),
    )

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        super()._parse_file(f, metadata, magic)
        metadata['~format'] = '%s (%s)' % (self.NAME, self._decode_text(magic))


class Composer669File(ModuleFile):
    EXTENSIONS = ['.669']
    NAME = 'Composer 669'

    # http://www.textfiles.com/programming/FORMATS/669-form.pro
    _magic_bytes = (
        MagicBytes(b'if'),
        MagicBytes(b'JN'),  # Extended 669
    )
    _static_text_fields = (
        StaticField('comment', 2, 108, FieldAccess.READ_WRITE),
    )

    def _parse_file(self, f: RawIOBase, metadata: Metadata, magic: MagicBytes):
        super()._parse_file(f, metadata, magic)
        if magic == b'JN':
            metadata['~format'] = 'Extended Composer 669'


class OktalyzerFile(ModuleFile):
    EXTENSIONS = ['.okt']
    NAME = 'Oktalyzer'

    # http://www.vgmpf.com/Wiki/index.php?title=OKT
    _magic_bytes = (MagicBytes(b'OKTASONGCMOD'),)


register_format(MODFile)
register_format(ExtendedModuleFile)
register_format(ImpulseTrackerFile)
register_format(AHXFile)
register_format(MEDFile)
register_format(MTMFile)
register_format(S3MFile)
register_format(ULTFile)
register_format(Composer669File)
register_format(OktalyzerFile)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/deezerart/options.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(420, 320)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.sizeLabel = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizeLabel.sizePolicy().hasHeightForWidth())
        self.sizeLabel.setSizePolicy(sizePolicy)
        self.sizeLabel.setObjectName("sizeLabel")
        self.gridLayout.addWidget(self.sizeLabel, 0, 0, 1, 1)
        self.size = QtWidgets.QComboBox(Form)
        self.size.setObjectName("size")
        self.gridLayout.addWidget(self.size, 0, 2, 1, 1)
        self.min_similarity_label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.min_similarity_label.sizePolicy().hasHeightForWidth())
        self.min_similarity_label.setSizePolicy(sizePolicy)
        self.min_similarity_label.setObjectName("min_similarity_label")
        self.gridLayout.addWidget(self.min_similarity_label, 1, 0, 1, 2)
        self.min_similarity = QtWidgets.QSpinBox(Form)
        self.min_similarity.setMaximum(100)
        self.min_similarity.setObjectName("min_similarity")
        self.gridLayout.addWidget(self.min_similarity, 1, 2, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 2, 1, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.sizeLabel.setText(_translate("Form", "Cover size:"))
        self.min_similarity_label.setText(_translate("Form", "Minimal similarity for matches:"))
        self.min_similarity.setSuffix(_translate("Form", " %"))
        self.min_similarity.setPrefix(_translate("Form", " "))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '.\plugins\genre_mapper\options_genre_mapper.ui'
#
# Created by: PyQt5 UI code generator 5.14.1
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GenreMapperOptionsPage(object):
    def setupUi(self, GenreMapperOptionsPage):
        GenreMapperOptionsPage.setObjectName("GenreMapperOptionsPage")
        GenreMapperOptionsPage.resize(398, 568)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(GenreMapperOptionsPage.sizePolicy().hasHeightForWidth())
        GenreMapperOptionsPage.setSizePolicy(sizePolicy)
        self.vboxlayout = QtWidgets.QVBoxLayout(GenreMapperOptionsPage)
        self.vboxlayout.setSpacing(16)
        self.vboxlayout.setObjectName("vboxlayout")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gm_description = QtWidgets.QGroupBox(GenreMapperOptionsPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.gm_description.sizePolicy().hasHeightForWidth())
        self.gm_description.setSizePolicy(sizePolicy)
        self.gm_description.setMinimumSize(QtCore.QSize(0, 50))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.gm_description.setFont(font)
        self.gm_description.setObjectName("gm_description")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.gm_description)
        self.verticalLayout.setContentsMargins(9, 9, 9, 1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.format_description = QtWidgets.QLabel(self.gm_description)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.format_description.sizePolicy().hasHeightForWidth())
        self.format_description.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.format_description.setFont(font)
        self.format_description.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.format_description.setWordWrap(True)
        self.format_description.setObjectName("format_description")
        self.verticalLayout.addWidget(self.format_description)
        self.verticalLayout_4.addWidget(self.gm_description)
        self.cb_enable_genre_mapping = QtWidgets.QCheckBox(GenreMapperOptionsPage)
        self.cb_enable_genre_mapping.setObjectName("cb_enable_genre_mapping")
        self.verticalLayout_4.addWidget(self.cb_enable_genre_mapping)
        self.gm_replacement_pairs = QtWidgets.QGroupBox(GenreMapperOptionsPage)
        self.gm_replacement_pairs.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.gm_replacement_pairs.sizePolicy().hasHeightForWidth())
        self.gm_replacement_pairs.setSizePolicy(sizePolicy)
        self.gm_replacement_pairs.setMinimumSize(QtCore.QSize(0, 50))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.gm_replacement_pairs.setFont(font)
        self.gm_replacement_pairs.setObjectName("gm_replacement_pairs")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.gm_replacement_pairs)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.cb_use_regex = QtWidgets.QCheckBox(self.gm_replacement_pairs)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.cb_use_regex.setFont(font)
        self.cb_use_regex.setObjectName("cb_use_regex")
        self.verticalLayout_3.addWidget(self.cb_use_regex)
        self.genre_mapper_first_match_only = QtWidgets.QCheckBox(self.gm_replacement_pairs)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.genre_mapper_first_match_only.setFont(font)
        self.genre_mapper_first_match_only.setChecked(False)
        self.genre_mapper_first_match_only.setObjectName("genre_mapper_first_match_only")
        self.verticalLayout_3.addWidget(self.genre_mapper_first_match_only)
        self.genre_mapper_replacement_pairs = QtWidgets.QPlainTextEdit(self.gm_replacement_pairs)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.genre_mapper_replacement_pairs.sizePolicy().hasHeightForWidth())
        self.genre_mapper_replacement_pairs.setSizePolicy(sizePolicy)
        self.genre_mapper_replacement_pairs.setMinimumSize(QtCore.QSize(0, 50))
        font = QtGui.QFont()
        font.setFamily("Courier New")
        font.setPointSize(10)
        self.genre_mapper_replacement_pairs.setFont(font)
        self.genre_mapper_replacement_pairs.setTabChangesFocus(True)
        self.genre_mapper_replacement_pairs.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.genre_mapper_replacement_pairs.setObjectName("genre_mapper_replacement_pairs")
        self.verticalLayout_3.addWidget(self.genre_mapper_replacement_pairs)
        self.verticalLayout_4.addWidget(self.gm_replacement_pairs)
        self.vboxlayout.addLayout(self.verticalLayout_4)

        self.retranslateUi(GenreMapperOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(GenreMapperOptionsPage)
        GenreMapperOptionsPage.setTabOrder(self.cb_enable_genre_mapping, self.cb_use_regex)
        GenreMapperOptionsPage.setTabOrder(self.cb_use_regex, self.genre_mapper_first_match_only)
        GenreMapperOptionsPage.setTabOrder(self.genre_mapper_first_match_only, self.genre_mapper_replacement_pairs)

    def retranslateUi(self, GenreMapperOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.gm_description.setTitle(_translate("GenreMapperOptionsPage", "Genre Mapper"))
        self.format_description.setText(_translate("GenreMapperOptionsPage", "<html><head/><body><p>These are the original / replacement pairs used to map one genre entry to another. Each pair must be entered on a separate line in the form:</p><p><span style=\" font-weight:600;\">[genre match test string]=[replacement genre]</span></p><p>Unless the &quot;regular expressions&quot; option is enabled, supported wildcards in the test string part of the mapping include \'*\' and \'?\' to match any number of characters and a single character respectively. An example for mapping all types of Rock genres (e.g. Country Rock, Hard Rock, Progressive Rock) to &quot;Rock&quot; would be done using the following line:</p><p><span style=\" font-family:\'Courier New\'; font-size:10pt; font-weight:600;\">*rock*=Rock</span></p><p>Blank lines and lines beginning with an equals sign (=) will be ignored. Case-insensitive tests are used when matching. Replacements will be made in the order they are found in the list.</p><p>For more information please see the <a href=\"https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/genre_mapper/docs/README.md\"><span style=\" text-decoration: underline; color:#0000ff;\">User Guide</span></a> on GitHub.<br/></p></body></html>"))
        self.cb_enable_genre_mapping.setText(_translate("GenreMapperOptionsPage", "Enable genre mapping"))
        self.gm_replacement_pairs.setTitle(_translate("GenreMapperOptionsPage", "Replacement Pairs"))
        self.cb_use_regex.setText(_translate("GenreMapperOptionsPage", "Match tests are entered as regular expressions"))
        self.genre_mapper_first_match_only.setText(_translate("GenreMapperOptionsPage", "Apply only the first matching replacement"))
        self.genre_mapper_replacement_pairs.setPlaceholderText(_translate("GenreMapperOptionsPage", "Enter replacement pairs (one per line)"))



from PyQt5 import QtCore, QtWidgets


class Ui_AddToCollectionOptions(object):
    def setupUi(self, AddToCollectionOptions):
        AddToCollectionOptions.setObjectName("AddToCollectionOptions")
        AddToCollectionOptions.resize(472, 215)
        self.verticalLayout = QtWidgets.QVBoxLayout(AddToCollectionOptions)
        self.verticalLayout.setObjectName("verticalLayout")
        self.collection_label = QtWidgets.QLabel(AddToCollectionOptions)
        self.collection_label.setObjectName("collection_label")
        self.verticalLayout.addWidget(self.collection_label)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed
        )
        self.collection_name = QtWidgets.QComboBox(AddToCollectionOptions)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.collection_name.sizePolicy().hasHeightForWidth()
        )
        self.collection_name.setSizePolicy(sizePolicy)
        self.collection_name.setEditable(False)
        self.collection_name.setObjectName("collection_name")
        self.verticalLayout.addWidget(self.collection_name)
        spacerItem = QtWidgets.QSpacerItem(
            20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        self.verticalLayout.addItem(spacerItem)

        self.retranslateUi(AddToCollectionOptions)
        QtCore.QMetaObject.connectSlotsByName(AddToCollectionOptions)

    def retranslateUi(self, AddToCollectionOptions):
        _translate = QtCore.QCoreApplication.translate
        AddToCollectionOptions.setWindowTitle(_("Form"))
        self.collection_label.setText(_("Collection to add releases to:"))
